#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from ctypes import (
    LittleEndianStructure,
    _SimpleCData,
    sizeof,
    c_double,
    c_int8,
    c_int32,
    c_uint8,
    c_uint16,
    c_uint32,
)
from typing import List, Dict, Sequence, Set, Tuple, Optional, Union, Any
from dataclasses import dataclass
from enum import Enum


class OperandType:
    name: str
    c_type: _SimpleCData

    def __init__(self, name: str, c_type: str):

        self.name = name
        self.c_type = {
            "int8_t": c_int8,
            "int32_t": c_int32,
            "uint8_t": c_uint8,
            "uint16_t": c_uint16,
            "uint32_t": c_uint32,
            "double": c_double,
        }[c_type]


class OperandMeaning(Enum):
    bigint_id = 1
    function_id = 2
    string_id = 3


@dataclass
class Operand:
    operand_type: OperandType
    operand_meaning: Optional[OperandMeaning]


class Instruction:
    name: str
    opcode: int
    operands: List[Operand]
    structure: LittleEndianStructure
    binary_size: int
    has_ret_target: bool

    def __init__(
        self, name: str, opcode: int, args: List[OperandType], module_ref: "module"
    ):

        self.name = name
        self.opcode = opcode
        self.operands = [Operand(arg, None) for arg in args]
        self.has_ret_target = False

        module_ref["_instructions"].append(self)

        self.structure = type(
            name,
            (LittleEndianStructure,),
            dict(
                _pack_=True,
                _fields_=[
                    ("arg%d" % (pos + 1), operand.c_type)
                    for pos, operand in enumerate(args)
                ],
            ),
        )

        self.binary_size = 1 + sizeof(self.structure)



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-

from collections import OrderedDict


"""
    Turn a parsed C structure in a dict of human-readable key-value
    pairs, for displayal in ASCII tables
    
    :param ctypes_structure: A parsed ctypes structures to consume.
    
    :returns An OrderedDict of strings/strings.
"""


def structure_to_key_values_strings(ctypes_structure):

    key_values = OrderedDict()

    if not getattr(ctypes_structure, "_fields_", None):
        return key_values

    for key, ctype in (field[:2] for field in ctypes_structure._fields_):
        value = getattr(ctypes_structure, key)

        # Turn "key_name" into "Key name"

        pretty_key = key[0].upper() + key[1:]
        pretty_key = pretty_key.replace("_", " ")

        # Stringify the value

        if type(value) == bytes:  # Strings
            key_values[pretty_key] = value.decode("ascii")

        elif type(value) == int:  # Integer
            # key_values[pretty_key] = '0x%08x' % value if value else 'N/A'
            key_values[pretty_key] = str(value) if value < (1 << 15) else hex(value)

        else:
            key_values[pretty_key] = bytes(value).hex()

    return key_values


"""
    Return an ASCII table from a parsed C structure, with field names as
    column 1 and values as column 2.
"""


def pretty_print_structure(ctypes_structure):

    key_values = structure_to_key_values_strings(ctypes_structure)

    pretty_print_table(list(key_values.items()))


"""
    Return an ascii table from a list (rows) of list (columns) of strings (cells)
"""


def pretty_print_table(rows):

    # Calculate columns length

    if not rows:
        return

    number_of_columns = len(rows[0])

    column_to_max_length = [
        max(len(row[column]) for row in rows) for column in range(number_of_columns)
    ]

    # Do a nice table

    print()

    print("+-%s-+" % "---".join("-" * max_len for max_len in column_to_max_length))

    for row in rows:
        print(
            "| %s |"
            % " | ".join(
                row[column].ljust(column_to_max_length[column])
                for column in range(number_of_columns)
            )
        )

        print("+-%s-+" % "---".join("-" * max_len for max_len in column_to_max_length))



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from typing import List, Dict, Set, Sequence, Union, Optional, Iterator, Any
from logging import warning
from io import BytesIO

# Imports relative to the current directory:
from hbc_opcodes import (
    hbc51,
    hbc58,
    hbc59,
    hbc61,
    hbc62,
    hbc68,
    hbc69,
    hbc70,
    hbc72,
    hbc73,
    hbc76,
    hbc80,
    hbc81,
    hbc82,
    hbc83,
    hbc84,
    hbc85,
    hbc86,
    hbc87,
    hbc89,
    hbc90,
    hbc92,
    hbc95,
)
from serialized_literal_parser import unpack_slp_array, SLPArray, SLPValue, TagType
from hbc_opcodes.def_classes import OperandMeaning, Instruction


class ParsedInstruction:
    inst: Instruction
    arg1: object
    arg2: object
    arg3: object
    arg4: object
    arg5: object
    arg6: object
    switch_jump_table: Optional[List[int]]
    original_pos: int
    next_pos: int
    hbc_reader: "HBCReader"

    def __repr__(self):
        operands = [
            "%s: %s"
            % (
                (
                    self.inst.operands[index].operand_meaning.name
                    if self.inst.operands[index].operand_meaning
                    else self.inst.operands[index].operand_type.name
                ),
                getattr(self, "arg%d" % (index + 1)),
            )
            for index in range(len(self.inst.operands))
        ]

        comment = ""
        for operand_index, operand in enumerate(self.inst.operands):
            if operand.operand_meaning:
                operand_value = getattr(self, "arg%d" % (operand_index + 1))
                if operand.operand_meaning == OperandMeaning.string_id:
                    comment += "  # String: %r (%s)" % (
                        self.hbc_reader.strings[operand_value],
                        self.hbc_reader.string_kinds[operand_value].name,
                    )
                elif operand.operand_meaning == OperandMeaning.bigint_id:
                    comment += (
                        "  # BigInt: %s"
                        % (self.hbc_reader.bigint_values[operand_value])
                    )
                elif operand.operand_meaning == OperandMeaning.function_id:
                    function_header = self.hbc_reader.function_headers[operand_value]
                    comment += (
                        "  # Function: [#%d %s of %d bytes]: %d params @ offset 0x%08x"
                        % (
                            operand_value,
                            self.hbc_reader.strings[function_header.functionName],
                            function_header.bytecodeSizeInBytes,
                            function_header.paramCount,
                            function_header.offset,
                        )
                    )
            elif operand.operand_type.name in ("Addr8", "Addr32"):
                operand_value = getattr(self, "arg%d" % (operand_index + 1))
                comment += "  # Address: %08x" % (self.original_pos + operand_value)
        if self.inst.name in ("NewArrayWithBuffer", "NewArrayWithBufferLong"):
            comment += "  # Array: [%s]" % ", ".join(
                unpack_slp_array(
                    self.hbc_reader.arrays[self.arg4 :], self.arg3
                ).to_strings(self.hbc_reader.strings)
            )
        elif self.inst.name in ("NewObjectWithBuffer", "NewObjectWithBufferLong"):
            comment += "  # Object: {%s}" % ", ".join(
                "%s: %s" % (key, value)
                for key, value in zip(
                    unpack_slp_array(
                        self.hbc_reader.object_keys[self.arg4 :], self.arg3
                    ).to_strings(self.hbc_reader.strings),
                    unpack_slp_array(
                        self.hbc_reader.object_values[self.arg5 :], self.arg3
                    ).to_strings(self.hbc_reader.strings),
                )
            )
        elif self.inst.name in ("CallBuiltin", "CallBuiltinLong", "GetBuiltinClosure"):
            builtin_number = self.arg2
            builtin_functions = get_builtin_functions(self.hbc_reader.parser_module)
            comment += "  # Built-in function: [#%d %s]" % (
                builtin_number,
                builtin_functions[builtin_number],
            )
        elif self.inst.name == "SwitchImm":
            comment += "  # Jump table: [%s]" % ", ".join(
                "%08x" % value for value in self.switch_jump_table
            )

        return f"{'%08x' % self.original_pos}: <{self.inst.name}>: <{', '.join(operands)}>{comment}"


def get_parser(bytecode_version: int) -> "module":
    parser_module_tbl = {
        51: hbc51,
        58: hbc58,
        59: hbc59,
        61: hbc61,
        62: hbc62,
        68: hbc68,
        69: hbc69,
        70: hbc70,
        # Before version 72/0.4.0, the bytecode format used predefined
        # string identifiers (which were removed in commit 63fb517)
        # These are not supported in the current code setup
        72: hbc72,
        73: hbc73,
        76: hbc76,
        80: hbc80,
        81: hbc81,
        82: hbc82,
        83: hbc83,
        84: hbc84,
        85: hbc85,
        86: hbc86,
        87: hbc87,
        89: hbc89,
        90: hbc90,
        92: hbc92,
        # The changes introduced in Hermes bytecode version 92
        # were reverted to the state of version 91 after a few
        # days from their introduction in the Git tree, into a
        # new version 93, and reintroduced in version 94.
        93: hbc90,
        94: hbc92,
        95: hbc95,
    }

    if bytecode_version < 72:
        warning(
            "This file uses an ancient Hermes bytecode format, which "
            + "is not supported."
        )

    elif bytecode_version > 96:
        warning(
            (
                "Bytecode version %d corresponds to a development or "
                + "recent version of the Hermes bytecode and is not "
                + "formally supported by the current tool."
            )
            % bytecode_version
        )

    for min_version in reversed(sorted(parser_module_tbl)):
        if bytecode_version >= min_version:
            parser_module = parser_module_tbl[min_version]
            break

    return parser_module


def get_builtin_functions(parser_module: "module") -> List[str]:

    return parser_module._builtin_function_names


# Read and parse the actual bytecode (+ switch tables) for
# a given function header


def parse_hbc_bytecode(
    function_header: object, hbc_reader: "HBCReader"
) -> Iterator[Instruction]:

    hbc_reader.file_buffer.seek(function_header.offset)

    buf = BytesIO(hbc_reader.file_buffer.read(function_header.bytecodeSizeInBytes))

    while True:
        original_pos = buf.tell()

        opcode = buf.read(1)
        if not opcode:
            break
        opcode = opcode[0]

        inst = hbc_reader.parser_module._opcode_to_instruction.get(opcode)
        if not inst:
            raise NotImplementedError(
                'Opcode "0x%02x" is currently not implemented within hermes-dec.'
                % opcode
            )

        structure = inst.structure()
        buf.readinto(structure)

        result = ParsedInstruction()
        result.inst = inst
        result.original_pos = original_pos
        result.next_pos = original_pos + inst.binary_size
        result.hbc_reader = hbc_reader

        for operand in ("arg1", "arg2", "arg3", "arg4", "arg5", "arg6"):
            if hasattr(structure, operand):
                setattr(result, operand, getattr(structure, operand))

        if inst.name == "SwitchImm":
            result.switch_jump_table = []
            hbc_reader.file_buffer.seek(
                function_header.offset + original_pos + structure.arg2
            )
            hbc_reader.align_over_padding()

            for jump_table_entry in range(structure.arg4, structure.arg5 + 1):
                result.switch_jump_table.append(
                    int.from_bytes(hbc_reader.file_buffer.read(4), "little")
                    + original_pos
                )

        yield result



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from typing import List, Tuple, Dict, Set, Sequence, Union, Optional, Any
from os.path import dirname, realpath

from defs import (
    HermesDecompiler,
    DecompiledFunctionBody,
    TokenString,
    LeftParenthesisToken,
    RightParenthesisToken,
    LeftHandRegToken,
    AssignmentToken,
    DotAccessorToken,
    CatchBlockStart,
    RightHandRegToken,
    GetEnvironmentToken,
    StoreToEnvironment,
    BindToken,
    SwitchImm,
    LoadFromEnvironmentToken,
    ReturnDirective,
    ResumeGenerator,
    SaveGenerator,
    StartGenerator,
    NewInnerEnvironmentToken,
    NewEnvironmentToken,
    JumpCondition,
    JumpNotCondition,
    FunctionTableIndex,
    ForInLoopInit,
    ForInLoopNextIter,
    ThrowDirective,
    ReturnDirective,
    RawToken,
)
from serialized_literal_parser import unpack_slp_array, SLPArray, SLPValue, TagType
from hbc_bytecode_parser import parse_hbc_bytecode


def pass2_transform_code(
    state: HermesDecompiler, function_body: DecompiledFunctionBody
):

    TS = TokenString

    RD = ReturnDirective
    BIND = BindToken

    RG = ResumeGenerator
    SG = SaveGenerator
    SG2 = StartGenerator

    LHRT = LeftHandRegToken
    AT = AssignmentToken
    RHRT = RightHandRegToken
    RT = RawToken

    GET = GetEnvironmentToken
    STE = StoreToEnvironment
    NET = NewEnvironmentToken
    NIET = NewInnerEnvironmentToken
    LFET = LoadFromEnvironmentToken

    TD = ThrowDirective
    RD = ReturnDirective

    SI = SwitchImm
    JC = JumpCondition
    JNC = JumpNotCondition

    FILI = ForInLoopInit
    FILNI = ForInLoopNextIter

    DAT = DotAccessorToken
    CBS = CatchBlockStart

    """
    BFT = BuiltinFunctionToken
    ACT = AsyncClosureToken
    CT = ClosureToken
    FT = FunctionToken
    """
    FTI = FunctionTableIndex

    LPT = LeftParenthesisToken
    RPT = RightParenthesisToken

    lines = function_body.statements = []

    function_header = function_body.function_object

    for instruction in parse_hbc_bytecode(function_header, state.hbc_reader):
        op1 = getattr(instruction, "arg1", None)
        op2 = getattr(instruction, "arg2", None)
        op3 = getattr(instruction, "arg3", None)
        op4 = getattr(instruction, "arg4", None)
        op5 = getattr(instruction, "arg5", None)
        op6 = getattr(instruction, "arg6", None)

        if instruction.inst.name in ("Add", "AddN"):
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" + "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Add32":
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT("__uasm.add32"),
                        LPT(),
                        RHRT(op2),
                        RT(", "),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "AddEmptyString":
            lines.append(
                TS([LHRT(op1), AT(), RT("'' + "), RHRT(op2)], assembly=[instruction])
            )
        elif instruction.inst.name == "AsyncBreakCheck":
            lines.append(TS([], assembly=[instruction]))
        elif instruction.inst.name == "BitAnd":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" & "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "BitNot":
            lines.append(
                TS([LHRT(op1), AT(), RT("~"), RHRT(op2)], assembly=[instruction])
            )
        elif instruction.inst.name == "BitOr":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" | "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "BitXor":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" ^ "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        # Note: It should be checked later whether the implementation of
        # parsing the "Call*" opcodes present below matches the difference
        # between the function_header.frameSize and the latest frame
        # register index across all version of the Hermes bytecode VM,
        # whether thisArg is handled correctly in all implementation of
        # the variants of the call, we should also eventually set up
        # unit tests, etc.
        elif instruction.inst.name in ("Call", "CallLong"):
            args = []
            for register in reversed(
                range(
                    function_header.frameSize - 7 - op3, function_header.frameSize - 7
                )
            ):
                args += [RHRT(register), RT(", ")]
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RHRT(function_header.frameSize - 7),
                        RT("["),
                        RHRT(op2),
                        RT("]"),
                        LPT(),
                        *args[:-1],
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("Call1", "Call2", "Call3", "Call4"):
            args = []
            for arg_count in range(1, int(instruction.inst.name[-1])):
                args += [RHRT([op4, op5, op6][arg_count - 1]), RT(", ")]
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), BIND(op3), LPT(), *args[:-1], RPT()],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("CallDirect", "CallDirectLongIndex"):
            args = []
            for register in reversed(
                range(
                    function_header.frameSize - 7 - op2, function_header.frameSize - 7
                )
            ):
                args += [RHRT(register), RT(", ")]
            state.calldirect_function_ids.add(op3)
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        DAT(),
                        FTI(op3, state),
                        BIND(function_header.frameSize - 7),
                        LPT(),
                        *args[:-1],
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("CallBuiltin", "CallBuiltinLong"):
            args = []
            for register in reversed(
                range(
                    function_header.frameSize - 7 - (op3 - 1),
                    function_header.frameSize - 7,
                )
            ):
                args += [RHRT(register), RT(", ")]
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        FTI(op2, state, is_builtin=True),
                        LPT(),
                        *args[:-1],
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Catch":
            lines.append(TS([CBS(op1)], assembly=[instruction]))
        elif instruction.inst.name == "CoerceThisNS":
            lines.append(TS([LHRT(op1), AT(), RHRT(op2)], assembly=[instruction]))
        elif instruction.inst.name == "CompleteGenerator":
            lines.append(TS([], assembly=[instruction]))
        elif instruction.inst.name in ("Construct", "ConstructLong"):
            args = []
            for register in reversed(
                range(
                    function_header.frameSize - 7 - op3, function_header.frameSize - 7
                )
            ):
                args += [RHRT(register), RT(", ")]
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT("new "),
                        RHRT(function_header.frameSize - 7),
                        RT("["),
                        RHRT(op2),
                        RT("]"),
                        LPT(),
                        *args[:-1],
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            "CreateAsyncClosure",
            "CreateAsyncClosureLongIndex",
        ):
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        FTI(op3, state, op2, is_async=True, is_closure=True),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("CreateClosure", "CreateClosureLongIndex"):
            lines.append(
                TS(
                    [LHRT(op1), AT(), FTI(op3, state, op2, is_closure=True)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "CreateEnvironment":
            lines.append(TS([NET(op1)], assembly=[instruction]))
        elif instruction.inst.name == "CreateInnerEnvironment":
            lines.append(TS([NIET(op1, op2, op3)], assembly=[instruction]))
        elif instruction.inst.name in ("CreateGenerator", "CreateGeneratorLongIndex"):
            lines.append(
                TS(
                    [LHRT(op1), AT(), FTI(op3, state, op2, is_generator=True)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            "CreateGeneratorClosure",
            "CreateGeneratorClosureLongIndex",
        ):
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        FTI(op3, state, op2, is_closure=True, is_generator=True),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "CreateRegExp":
            pattern_string = state.hbc_reader.strings[op2]
            flags_string = state.hbc_reader.strings[op3]

            pattern_string = pattern_string.translate(
                {char: repr(char) for char in range(0x20)}
            )
            pattern_string = pattern_string.replace("/", "\\/")

            readable_regexp = "/%s/%s" % (pattern_string, flags_string)

            lines.append(
                TS([LHRT(op1), AT(), RT(readable_regexp)], assembly=[instruction])
            )
        elif instruction.inst.name == "CreateThis":
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT("Object.create"),
                        LPT(),
                        RHRT(op2),
                        RT(", {constructor: {value: "),
                        RHRT(op3),
                        RT("}}"),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Debugger":
            lines.append(TS([RT("debugger")], assembly=[instruction]))
        elif instruction.inst.name == "DebuggerBreakCheck":
            lines.append(TS([], assembly=[instruction]))
        elif instruction.inst.name == "Dec":
            lines.append(
                TS([LHRT(op1), AT(), RHRT(op2), RT(" - 1")], assembly=[instruction])
            )
        elif instruction.inst.name == "DeclareGlobalVar":
            global_var = state.hbc_reader.strings[op1]

            lines.append(
                TS([RT(global_var), AT(), RT("undefined")], assembly=[instruction])
            )
        elif instruction.inst.name in ("DelById", "DelByIdLong"):
            prop_string = state.hbc_reader.strings[op3]

            lines.append(
                TS(
                    [LHRT(op1), AT(), RT("delete "), RHRT(op2), DAT(), RT(prop_string)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "DelByVal":
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT("delete "),
                        RHRT(op2),
                        RT("["),
                        RHRT(op3),
                        RT("]"),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "DirectEval":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT("eval"), LPT(), RHRT(op2), RPT()],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("Div", "DivN"):
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" / "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Divi32":
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT("__uasm.divi32"),
                        LPT(),
                        RHRT(op2),
                        RT(", "),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Divu32":
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT("__uasm.divu32"),
                        LPT(),
                        RHRT(op2),
                        RT(", "),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Eq":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" == "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "GetArgumentsLength":
            lines.append(
                TS([LHRT(op1), AT(), RT("arguments.length")], assembly=[instruction])
            )
        elif instruction.inst.name == "GetArgumentsPropByVal":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT("arguments"), RT("["), RHRT(op2), RT("]")],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "GetBuiltinClosure":
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        FTI(op2, state, is_builtin=True, is_closure=True),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            "GetById",
            "GetByIdLong",
            "GetByIdShort",
            "TryGetById",
            "TryGetByIdLong",
        ):
            string = state.hbc_reader.strings[op4]

            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), DAT(), RT(string)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "GetByVal":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT("["), RHRT(op3), RT("]")],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "GetEnvironment":
            lines.append(TS([GET(op1, op2)], assembly=[instruction]))
        elif instruction.inst.name == "GetGlobalObject":
            lines.append(TS([LHRT(op1), AT(), RT("global")], assembly=[instruction]))
        elif instruction.inst.name == "GetNewTarget":
            lines.append(
                TS([LHRT(op1), AT(), RT("new.target")], assembly=[instruction])
            )
        elif instruction.inst.name == "GetNextPName":
            lines.append(TS([FILNI(op1, op2, op3, op4, op5)], assembly=[instruction]))
        elif instruction.inst.name == "GetPNameList":
            lines.append(TS([FILI(op1, op2, op3, op4)], assembly=[instruction]))
        elif instruction.inst.name == "Greater":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" > "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "GreaterEq":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" >= "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Inc":
            lines.append(
                TS([LHRT(op1), AT(), RHRT(op2), RT(" + 1")], assembly=[instruction])
            )
        elif instruction.inst.name == "InstanceOf":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" instanceof "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "IsIn":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" in "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        # Note: The implementations of the methods related to the iteration below
        # is not fully consistent with the corresponding React Native instruction
        # behavior, which has special handling for iterating over arrays, may
        # suppress propagating the exception over certain IteratorClose() calls...
        #
        # This may be improved in the future if any required.
        elif instruction.inst.name == "IteratorBegin":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT("[Symbol.iterator]")],
                    assembly=[instruction],
                )
            )
            lines.append(
                TS(
                    [LHRT(op2), AT(), RHRT(op1), LPT(), RPT(), DAT(), RT("next")],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "IteratorNext":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op3), LPT(), RPT(), DAT(), RT("value")],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "IteratorClose":
            lines.append(
                TS(
                    [RHRT(op1), DAT(), RT("return"), LPT(), RPT()],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("JEqual", "JEqualLong"):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" != "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" == "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            "JGreater",
            "JGreaterN",
            "JGreaterLong",
            "JGreaterNLong",
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RT("!"),
                            LPT(),
                            RHRT(op2),
                            RT(" > "),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" > "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            "JNotGreater",
            "JNotGreaterN",
            "JNotGreaterLong",
            "JNotGreaterNLong",
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" > "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RT("!"),
                            LPT(),
                            RHRT(op2),
                            RT(" > "),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            "JGreaterEqual",
            "JGreaterEqualN",
            "JGreaterEqualLong",
            "JGreaterEqualNLong",
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RT("!"),
                            LPT(),
                            RHRT(op2),
                            RT(" >= "),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" >= "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            "JNotGreaterEqual",
            "JNotGreaterEqualN",
            "JNotGreaterEqualLong",
            "JNotGreaterEqualNLong",
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" >= "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RT("!"),
                            LPT(),
                            RHRT(op2),
                            RT(" >= "),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ("JLess", "JLessN", "JLessLong", "JLessNLong"):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RT("!"),
                            LPT(),
                            RHRT(op2),
                            RT(" < "),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" < "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            "JNotLess",
            "JNotLessN",
            "JNotLessLong",
            "JNotLessNLong",
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" < "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RT("!"),
                            LPT(),
                            RHRT(op2),
                            RT(" < "),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            "JLessEqual",
            "JLessEqualN",
            "JLessEqualLong",
            "JLessEqualNLong",
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RT("!"),
                            LPT(),
                            RHRT(op2),
                            RT(" <= "),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" <= "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            "JNotLessEqual",
            "JNotLessEqualN",
            "JNotLessEqualLong",
            "JNotLessEqualNLong",
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" <= "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RT("!"),
                            LPT(),
                            RHRT(op2),
                            RT(" <= "),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ("JNotEqual", "JNotEqualLong"):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" == "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" != "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ("JStrictEqual", "JStrictEqualLong"):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" !== "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" === "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ("JStrictNotEqual", "JStrictNotEqualLong"):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" === "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" !== "),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ("Jmp", "JmpLong"):
            if op1 > 0:
                lines.append(
                    TS(
                        [JNC(instruction.original_pos + op1), RT("false")],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [JC(instruction.original_pos + op1), RT("true")],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ("JmpFalse", "JmpFalseLong"):
            if op1 > 0:
                lines.append(
                    TS(
                        [JNC(instruction.original_pos + op1), RHRT(op2)],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [JC(instruction.original_pos + op1), RT("!"), RHRT(op2)],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ("JmpTrue", "JmpTrueLong"):
            if op1 > 0:
                lines.append(
                    TS(
                        [JNC(instruction.original_pos + op1), RT("!"), RHRT(op2)],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [JC(instruction.original_pos + op1), RHRT(op2)],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ("JmpUndefined", "JmpUndefinedLong"):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" !== undefined"),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(" === undefined"),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name == "LShift":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" << "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Less":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" < "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "LessEq":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" <= "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("LoadConstBigInt", "LoadConstBigIntLongIndex"):
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT(str(state.hbc_reader.bigint_values[op2]) + "n"),
                    ],
                    assembly=[instruction],
                )
            )  # TODO : Decode BigInts correctly
        elif instruction.inst.name in (
            "LoadConstDouble",
            "LoadConstInt",
            "LoadConstUInt8",
        ):
            lines.append(TS([LHRT(op1), AT(), str(op2)], assembly=[instruction]))
        elif instruction.inst.name in ("LoadConstEmpty", "LoadConstUndefined"):
            lines.append(TS([LHRT(op1), AT(), RT("undefined")], assembly=[instruction]))
        elif instruction.inst.name == "LoadConstFalse":
            lines.append(TS([LHRT(op1), AT(), RT("false")], assembly=[instruction]))
        elif instruction.inst.name == "LoadConstNull":
            lines.append(TS([LHRT(op1), AT(), RT("null")], assembly=[instruction]))
        elif instruction.inst.name in ("LoadConstString", "LoadConstStringLongIndex"):
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT(repr(state.hbc_reader.strings[op2]))],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "LoadConstTrue":
            lines.append(TS([LHRT(op1), AT(), RT("true")], assembly=[instruction]))
        elif instruction.inst.name == "LoadConstZero":
            lines.append(TS([LHRT(op1), AT(), RT("0")], assembly=[instruction]))
        elif instruction.inst.name in ("LoadFromEnvironment", "LoadFromEnvironmentL"):
            lines.append(TS([LHRT(op1), AT(), LFET(op2, op3)], assembly=[instruction]))
        elif instruction.inst.name in ("LoadParam", "LoadParamLong"):
            if not op2:
                arg_name = "this"
            elif op2 < function_body.function_object.paramCount:
                arg_name = "a" + str(op2 - 1)
            else:
                arg_name = "arguments[%d]" % (op2 - 1)
            lines.append(TS([LHRT(op1), AT(), RT(arg_name)], assembly=[instruction]))
        elif instruction.inst.name == "LoadThisNS":
            lines.append(TS([LHRT(op1), AT(), RT("this")], assembly=[instruction]))
        elif instruction.inst.name in (
            "Loadi16",
            "Loadi32",
            "Loadi8",
            "Loadu16",
            "Loadu32",
            "Loadu8",
        ):
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT("__uasm." + instruction.inst.name.lower()),
                        LPT(),
                        RHRT(op2),
                        RT(", "),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Mod":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" % "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("Mov", "MovLong"):
            lines.append(TS([LHRT(op1), AT(), RHRT(op2)], assembly=[instruction]))
        elif instruction.inst.name == "Mul32":
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT("__uasm.mul32"),
                        LPT(),
                        RHRT(op2),
                        RT(", "),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("Mul", "MulN"):
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" * "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Negate":
            lines.append(
                TS([LHRT(op1), AT(), RT("-"), RHRT(op2)], assembly=[instruction])
            )
        elif instruction.inst.name == "Neq":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" != "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "NewArray":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT("new Array"), LPT(), RT(str(op2)), RPT()],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("NewArrayWithBuffer", "NewArrayWithBufferLong"):
            array_text = "[%s]" % ", ".join(
                unpack_slp_array(state.hbc_reader.arrays[op4:], op3).to_strings(
                    state.hbc_reader.strings
                )
            )
            lines.append(TS([LHRT(op1), AT(), RT(array_text)], assembly=[instruction]))
        elif instruction.inst.name == "NewObject":
            lines.append(TS([LHRT(op1), AT(), RT("{}")], assembly=[instruction]))
        elif instruction.inst.name in (
            "NewObjectWithBuffer",
            "NewObjectWithBufferLong",
        ):
            object_text = "{%s}" % ", ".join(
                "%s: %s" % (key, value)
                for key, value in zip(
                    unpack_slp_array(
                        state.hbc_reader.object_keys[op4:], op3
                    ).to_strings(state.hbc_reader.strings),
                    unpack_slp_array(
                        state.hbc_reader.object_values[op5:], op3
                    ).to_strings(state.hbc_reader.strings),
                )
            )
            lines.append(TS([LHRT(op1), AT(), RT(object_text)], assembly=[instruction]))
        elif instruction.inst.name == "NewObjectWithParent":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT("Object.create"), LPT(), RHRT(op2), RPT()],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Not":
            lines.append(
                TS([LHRT(op1), AT(), RT("!"), RHRT(op2)], assembly=[instruction])
            )
        elif instruction.inst.name == "ProfilePoint":
            lines.append(TS([], assembly=[instruction]))
        elif instruction.inst.name in (
            "PutById",
            "PutByIdLong",
            "TryPutById",
            "TryPutByIdLong",
        ):
            index = repr(state.hbc_reader.strings[op4])
            lines.append(
                TS(
                    [LHRT(op1), RT("[" + index + "]"), AT(), RHRT(op2)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            "PutNewOwnById",
            "PutNewOwnByIdLong",
            "PutNewOwnByIdShort",
        ):
            index = repr(state.hbc_reader.strings[op3])
            lines.append(
                TS(
                    [LHRT(op1), RT("[" + index + "]"), AT(), RHRT(op2)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("PutNewOwnNEById", "PutNewOwnNEByIdLong"):
            index = repr(state.hbc_reader.strings[op3])
            lines.append(
                TS(
                    [
                        RT("Object.defineProperty"),
                        LPT(),
                        LHRT(op1),
                        RT(", "),
                        RT(repr(index)),
                        RT(", {value: "),
                        RHRT(op2),
                        RT("}"),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
            # TODO: Are non-enumerable values set correctly?
            # When are these used if they are used?
        elif instruction.inst.name == "PutByVal":
            lines.append(
                TS(
                    [LHRT(op1), RT("["), RHRT(op2), RT("]"), AT(), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "PutOwnByVal":
            if op4:  # Is the property enumerable?
                lines.append(
                    TS(
                        [LHRT(op1), RT("["), RHRT(op3), RT("]"), AT(), RHRT(op2)],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            RT("Object.defineProperty"),
                            LPT(),
                            LHRT(op1),
                            RT(", "),
                            RHRT(op3),
                            RT(", {value: "),
                            RHRT(op2),
                            RT("}"),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ("PutOwnByIndex", "PutOwnByIndexL"):
            lines.append(
                TS(
                    [LHRT(op1), RT("[%d]" % op3), AT(), RHRT(op2)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("PutOwnGetterSetterByVal"):
            index = repr(state.hbc_reader.strings[op3])
            lines.append(
                TS(
                    [
                        RT("Object.defineProperty"),
                        LPT(),
                        LHRT(op1),
                        RT(", "),
                        RHRT(op2),
                        RT(", {get: "),
                        RHRT(op3),
                        RT(", set: "),
                        RHRT(op4),
                        RT(", enumerable: " + ("true" if op5 else "false") + "}"),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "RShift":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" >> "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "ReifyArguments":
            lines.append(TS([LHRT(op1), AT(), RT("arguments")], assembly=[instruction]))
        elif instruction.inst.name == "ResumeGenerator":
            lines.append(TS([RG(op1, op2)], assembly=[instruction]))
        elif instruction.inst.name == "Ret":
            lines.append(TS([RD(), RHRT(op1)], assembly=[instruction]))
        elif instruction.inst.name in ("SaveGenerator", "SaveGeneratorLong"):
            lines.append(
                TS([SG(instruction.original_pos + op1)], assembly=[instruction])
            )
        elif instruction.inst.name == "SelectObject":
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RHRT(op3),
                        RT(" instanceof Object ? "),
                        RHRT(op3),
                        RT(" : "),
                        RHRT(op2),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "StartGenerator":
            lines.append(TS([SG2()], assembly=[instruction]))
        elif instruction.inst.name in ("Store16", "Store32", "Store8"):
            lines.append(
                TS(
                    [
                        RT("__uasm." + instruction.inst.name.lower()),
                        LPT(),
                        RHRT(op1),
                        RT(", "),
                        RHRT(op2),
                        RT(", "),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            "StoreNPToEnvironment",
            "StoreNPToEnvironmentL",
            "StoreToEnvironment",
            "StoreToEnvironmentL",
        ):
            lines.append(TS([STE(op1, op2, op3)], assembly=[instruction]))
        elif instruction.inst.name == "StrictEq":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" === "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "StrictNeq":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" !== "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ("Sub", "SubN"):
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" - "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Sub32":
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT("__uasm.sub32"),
                        LPT(),
                        RHRT(op2),
                        RT(", "),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "SwitchImm":
            lines.append(
                TS(
                    [
                        SI(
                            op1,
                            instruction.original_pos + op2,
                            instruction.original_pos + op3,
                            op4,
                            op5,
                        ),
                        RT(" // Switch table: %s" % instruction.switch_jump_table),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Throw":
            lines.append(TS([TD(), RHRT(op1)], assembly=[instruction]))
        elif instruction.inst.name == "ThrowIfEmpty":
            lines.append(
                TS(
                    [
                        RT("if"),
                        LPT(),
                        RT("!"),
                        RHRT(op2),
                        RPT(),
                        RT(" "),
                        TD(),
                        RT('ReferenceError("accessing an uninitialized variable")'),
                    ],
                    assembly=[instruction],
                )
            )
            lines.append(TS([RT("else "), LHRT(op1), AT(), RHRT(op2)], assembly=[]))
        elif instruction.inst.name == "ThrowIfHasRestrictedGlobalProperty":
            global_var = state.hbc_reader.strings[op1]

            lines.append(
                TS(
                    [
                        RT("if"),
                        LPT(),
                        RT("typeof global"),
                        DAT(),
                        RT(global_var),
                        RT(" === 'undefined'"),
                        RPT(),
                        RT(" "),
                        TD(),
                        RT('SyntaxError("Name is a restricted global identifier")'),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "ThrowIfUndefinedInst":
            lines.append(
                TS(
                    [
                        RT("if"),
                        LPT(),
                        RT("typeof "),
                        RHRT(op1),
                        RT(" === 'undefined'"),
                        RPT(),
                        RT(" "),
                        TD(),
                        RT('ReferenceError("accessing an uninitialized variable")'),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "ToInt32":
            lines.append(
                TS([LHRT(op1), AT(), RHRT(op2), RT(" | 0")], assembly=[instruction])
            )
        elif instruction.inst.name == "ToNumber":
            lines.append(
                TS([LHRT(op1), AT(), RHRT(op2), RT(" - 0")], assembly=[instruction])
            )
        elif instruction.inst.name == "ToNumeric":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT("parseFloat"), LPT(), RHRT(op2), RPT()],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "TypeOf":
            lines.append(
                TS([LHRT(op1), AT(), RT("typeof "), RHRT(op2)], assembly=[instruction])
            )
        elif instruction.inst.name == "URshift":
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(" >>> "), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == "Unreachable":
            raise ValueError("Unreachable position reached: %r" % instruction)

        else:
            lines.append(
                TS(
                    [RT("// Unsupported instruction: %r" % instruction)],
                    assembly=[instruction],
                )
            )
            """
            # Todo: We should likely put an "Error comment" in the decompiled
            # source later, instead of this
            raise NotImplementedError('[Unsupported instruction in function "%s":] %r' % (
                function_body.function_name,
                instruction
            ))
            """



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
"""
Note: The contents of the current file have been automatically
generated by the "utils/hermes_bytecode_structs_parser.py"
script

Please do not edit it manually. 👍
"""

from typing import List, Set, Dict, Union, Optional, Sequence, Any

# Imports from the current diretory
from .def_classes import *

_instructions: List[Instruction] = []

Reg8 = OperandType("Reg8", "uint8_t")
Reg32 = OperandType("Reg32", "uint32_t")
UInt8 = OperandType("UInt8", "uint8_t")
UInt16 = OperandType("UInt16", "uint16_t")
UInt32 = OperandType("UInt32", "uint32_t")
Addr8 = OperandType("Addr8", "int8_t")
Addr32 = OperandType("Addr32", "int32_t")
Imm32 = OperandType("Imm32", "int32_t")
Double = OperandType("Double", "double")

NewObjectWithBuffer = Instruction(
    "NewObjectWithBuffer", 0, [Reg8, UInt16, UInt16, UInt16, UInt16], globals()
)

NewObjectWithBufferLong = Instruction(
    "NewObjectWithBufferLong", 1, [Reg8, UInt16, UInt16, UInt32, UInt32], globals()
)

NewObject = Instruction("NewObject", 2, [Reg8], globals())

NewObjectWithParent = Instruction("NewObjectWithParent", 3, [Reg8, Reg8], globals())

NewArrayWithBuffer = Instruction(
    "NewArrayWithBuffer", 4, [Reg8, UInt16, UInt16, UInt16], globals()
)

NewArrayWithBufferLong = Instruction(
    "NewArrayWithBufferLong", 5, [Reg8, UInt16, UInt16, UInt32], globals()
)

NewArray = Instruction("NewArray", 6, [Reg8, UInt16], globals())

Mov = Instruction("Mov", 7, [Reg8, Reg8], globals())

MovLong = Instruction("MovLong", 8, [Reg32, Reg32], globals())

Negate = Instruction("Negate", 9, [Reg8, Reg8], globals())

Not = Instruction("Not", 10, [Reg8, Reg8], globals())

BitNot = Instruction("BitNot", 11, [Reg8, Reg8], globals())

TypeOf = Instruction("TypeOf", 12, [Reg8, Reg8], globals())

Eq = Instruction("Eq", 13, [Reg8, Reg8, Reg8], globals())

StrictEq = Instruction("StrictEq", 14, [Reg8, Reg8, Reg8], globals())

Neq = Instruction("Neq", 15, [Reg8, Reg8, Reg8], globals())

StrictNeq = Instruction("StrictNeq", 16, [Reg8, Reg8, Reg8], globals())

Less = Instruction("Less", 17, [Reg8, Reg8, Reg8], globals())

LessEq = Instruction("LessEq", 18, [Reg8, Reg8, Reg8], globals())

Greater = Instruction("Greater", 19, [Reg8, Reg8, Reg8], globals())

GreaterEq = Instruction("GreaterEq", 20, [Reg8, Reg8, Reg8], globals())

Add = Instruction("Add", 21, [Reg8, Reg8, Reg8], globals())

AddN = Instruction("AddN", 22, [Reg8, Reg8, Reg8], globals())

Mul = Instruction("Mul", 23, [Reg8, Reg8, Reg8], globals())

MulN = Instruction("MulN", 24, [Reg8, Reg8, Reg8], globals())

Div = Instruction("Div", 25, [Reg8, Reg8, Reg8], globals())

DivN = Instruction("DivN", 26, [Reg8, Reg8, Reg8], globals())

Mod = Instruction("Mod", 27, [Reg8, Reg8, Reg8], globals())

Sub = Instruction("Sub", 28, [Reg8, Reg8, Reg8], globals())

SubN = Instruction("SubN", 29, [Reg8, Reg8, Reg8], globals())

LShift = Instruction("LShift", 30, [Reg8, Reg8, Reg8], globals())

RShift = Instruction("RShift", 31, [Reg8, Reg8, Reg8], globals())

URshift = Instruction("URshift", 32, [Reg8, Reg8, Reg8], globals())

BitAnd = Instruction("BitAnd", 33, [Reg8, Reg8, Reg8], globals())

BitXor = Instruction("BitXor", 34, [Reg8, Reg8, Reg8], globals())

BitOr = Instruction("BitOr", 35, [Reg8, Reg8, Reg8], globals())

InstanceOf = Instruction("InstanceOf", 36, [Reg8, Reg8, Reg8], globals())

IsIn = Instruction("IsIn", 37, [Reg8, Reg8, Reg8], globals())

GetEnvironment = Instruction("GetEnvironment", 38, [Reg8, UInt8], globals())

StoreToEnvironment = Instruction(
    "StoreToEnvironment", 39, [Reg8, UInt8, Reg8], globals()
)

StoreToEnvironmentL = Instruction(
    "StoreToEnvironmentL", 40, [Reg8, UInt16, Reg8], globals()
)

StoreNPToEnvironment = Instruction(
    "StoreNPToEnvironment", 41, [Reg8, UInt8, Reg8], globals()
)

StoreNPToEnvironmentL = Instruction(
    "StoreNPToEnvironmentL", 42, [Reg8, UInt16, Reg8], globals()
)

LoadFromEnvironment = Instruction(
    "LoadFromEnvironment", 43, [Reg8, Reg8, UInt8], globals()
)

LoadFromEnvironmentL = Instruction(
    "LoadFromEnvironmentL", 44, [Reg8, Reg8, UInt16], globals()
)

GetGlobalObject = Instruction("GetGlobalObject", 45, [Reg8], globals())

GetNewTarget = Instruction("GetNewTarget", 46, [Reg8], globals())

CreateEnvironment = Instruction("CreateEnvironment", 47, [Reg8], globals())

DeclareGlobalVar = Instruction("DeclareGlobalVar", 48, [UInt32], globals())

DeclareGlobalVar.operands[0].operand_meaning = OperandMeaning.string_id

GetByIdShort = Instruction("GetByIdShort", 49, [Reg8, Reg8, UInt8, UInt8], globals())

GetById = Instruction("GetById", 50, [Reg8, Reg8, UInt8, UInt16], globals())

GetByIdLong = Instruction("GetByIdLong", 51, [Reg8, Reg8, UInt8, UInt32], globals())

GetByIdShort.operands[3].operand_meaning = OperandMeaning.string_id

GetById.operands[3].operand_meaning = OperandMeaning.string_id

GetByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

TryGetById = Instruction("TryGetById", 52, [Reg8, Reg8, UInt8, UInt16], globals())

TryGetByIdLong = Instruction(
    "TryGetByIdLong", 53, [Reg8, Reg8, UInt8, UInt32], globals()
)

TryGetById.operands[3].operand_meaning = OperandMeaning.string_id

TryGetByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

PutById = Instruction("PutById", 54, [Reg8, Reg8, UInt8, UInt16], globals())

PutByIdLong = Instruction("PutByIdLong", 55, [Reg8, Reg8, UInt8, UInt32], globals())

PutById.operands[3].operand_meaning = OperandMeaning.string_id

PutByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

TryPutById = Instruction("TryPutById", 56, [Reg8, Reg8, UInt8, UInt16], globals())

TryPutByIdLong = Instruction(
    "TryPutByIdLong", 57, [Reg8, Reg8, UInt8, UInt32], globals()
)

TryPutById.operands[3].operand_meaning = OperandMeaning.string_id

TryPutByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

PutNewOwnByIdShort = Instruction(
    "PutNewOwnByIdShort", 58, [Reg8, Reg8, UInt8], globals()
)

PutNewOwnById = Instruction("PutNewOwnById", 59, [Reg8, Reg8, UInt16], globals())

PutNewOwnByIdLong = Instruction(
    "PutNewOwnByIdLong", 60, [Reg8, Reg8, UInt32], globals()
)

PutNewOwnByIdShort.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnById.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnByIdLong.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnNEById = Instruction("PutNewOwnNEById", 61, [Reg8, Reg8, UInt16], globals())

PutNewOwnNEByIdLong = Instruction(
    "PutNewOwnNEByIdLong", 62, [Reg8, Reg8, UInt32], globals()
)

PutNewOwnNEById.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnNEByIdLong.operands[2].operand_meaning = OperandMeaning.string_id

PutOwnByIndex = Instruction("PutOwnByIndex", 63, [Reg8, Reg8, UInt8], globals())

PutOwnByIndexL = Instruction("PutOwnByIndexL", 64, [Reg8, Reg8, UInt32], globals())

PutOwnByVal = Instruction("PutOwnByVal", 65, [Reg8, Reg8, Reg8, UInt8], globals())

DelById = Instruction("DelById", 66, [Reg8, Reg8, UInt16], globals())

DelByIdLong = Instruction("DelByIdLong", 67, [Reg8, Reg8, UInt32], globals())

DelById.operands[2].operand_meaning = OperandMeaning.string_id

DelByIdLong.operands[2].operand_meaning = OperandMeaning.string_id

GetByVal = Instruction("GetByVal", 68, [Reg8, Reg8, Reg8], globals())

PutByVal = Instruction("PutByVal", 69, [Reg8, Reg8, Reg8], globals())

DelByVal = Instruction("DelByVal", 70, [Reg8, Reg8, Reg8], globals())

PutOwnGetterSetterByVal = Instruction(
    "PutOwnGetterSetterByVal", 71, [Reg8, Reg8, Reg8, Reg8, UInt8], globals()
)

GetPNameList = Instruction("GetPNameList", 72, [Reg8, Reg8, Reg8, Reg8], globals())

GetNextPName = Instruction(
    "GetNextPName", 73, [Reg8, Reg8, Reg8, Reg8, Reg8], globals()
)

Call = Instruction("Call", 74, [Reg8, Reg8, UInt8], globals())
Call.has_ret_target = True

Construct = Instruction("Construct", 75, [Reg8, Reg8, UInt8], globals())
Construct.has_ret_target = True

Call1 = Instruction("Call1", 76, [Reg8, Reg8, Reg8], globals())
Call1.has_ret_target = True

CallDirect = Instruction("CallDirect", 77, [Reg8, UInt8, UInt16], globals())
CallDirect.has_ret_target = True

Call2 = Instruction("Call2", 78, [Reg8, Reg8, Reg8, Reg8], globals())
Call2.has_ret_target = True

Call3 = Instruction("Call3", 79, [Reg8, Reg8, Reg8, Reg8, Reg8], globals())
Call3.has_ret_target = True

Call4 = Instruction("Call4", 80, [Reg8, Reg8, Reg8, Reg8, Reg8, Reg8], globals())
Call4.has_ret_target = True

CallLong = Instruction("CallLong", 81, [Reg8, Reg8, UInt32], globals())
CallLong.has_ret_target = True

ConstructLong = Instruction("ConstructLong", 82, [Reg8, Reg8, UInt32], globals())
ConstructLong.has_ret_target = True

CallDirectLongIndex = Instruction(
    "CallDirectLongIndex", 83, [Reg8, UInt8, UInt32], globals()
)
CallDirectLongIndex.has_ret_target = True

CallBuiltin = Instruction("CallBuiltin", 84, [Reg8, UInt8, UInt8], globals())

Ret = Instruction("Ret", 85, [Reg8], globals())

Catch = Instruction("Catch", 86, [Reg8], globals())

DirectEval = Instruction("DirectEval", 87, [Reg8, Reg8], globals())

Throw = Instruction("Throw", 88, [Reg8], globals())

Debugger = Instruction("Debugger", 89, [], globals())

DebuggerCheckBreak = Instruction("DebuggerCheckBreak", 90, [], globals())

ProfilePoint = Instruction("ProfilePoint", 91, [UInt16], globals())

Unreachable = Instruction("Unreachable", 92, [], globals())

CreateClosure = Instruction("CreateClosure", 93, [Reg8, Reg8, UInt16], globals())

CreateClosureLongIndex = Instruction(
    "CreateClosureLongIndex", 94, [Reg8, Reg8, UInt32], globals()
)

CreateGeneratorClosure = Instruction(
    "CreateGeneratorClosure", 95, [Reg8, Reg8, UInt16], globals()
)

CreateGeneratorClosureLongIndex = Instruction(
    "CreateGeneratorClosureLongIndex", 96, [Reg8, Reg8, UInt32], globals()
)

CreateThis = Instruction("CreateThis", 97, [Reg8, Reg8, Reg8], globals())

SelectObject = Instruction("SelectObject", 98, [Reg8, Reg8, Reg8], globals())

LoadParam = Instruction("LoadParam", 99, [Reg8, UInt8], globals())

LoadParamLong = Instruction("LoadParamLong", 100, [Reg8, UInt32], globals())

LoadConstUInt8 = Instruction("LoadConstUInt8", 101, [Reg8, UInt8], globals())

LoadConstInt = Instruction("LoadConstInt", 102, [Reg8, Imm32], globals())

LoadConstDouble = Instruction("LoadConstDouble", 103, [Reg8, Double], globals())

LoadConstString = Instruction("LoadConstString", 104, [Reg8, UInt16], globals())

LoadConstStringLongIndex = Instruction(
    "LoadConstStringLongIndex", 105, [Reg8, UInt32], globals()
)

LoadConstString.operands[1].operand_meaning = OperandMeaning.string_id

LoadConstStringLongIndex.operands[1].operand_meaning = OperandMeaning.string_id

LoadConstUndefined = Instruction("LoadConstUndefined", 106, [Reg8], globals())

LoadConstNull = Instruction("LoadConstNull", 107, [Reg8], globals())

LoadConstTrue = Instruction("LoadConstTrue", 108, [Reg8], globals())

LoadConstFalse = Instruction("LoadConstFalse", 109, [Reg8], globals())

LoadConstZero = Instruction("LoadConstZero", 110, [Reg8], globals())

CoerceThisNS = Instruction("CoerceThisNS", 111, [Reg8, Reg8], globals())

LoadThisNS = Instruction("LoadThisNS", 112, [Reg8], globals())

ToNumber = Instruction("ToNumber", 113, [Reg8, Reg8], globals())

ToInt32 = Instruction("ToInt32", 114, [Reg8, Reg8], globals())

AddEmptyString = Instruction("AddEmptyString", 115, [Reg8, Reg8], globals())

GetArgumentsPropByVal = Instruction(
    "GetArgumentsPropByVal", 116, [Reg8, Reg8, Reg8], globals()
)

GetArgumentsLength = Instruction("GetArgumentsLength", 117, [Reg8, Reg8], globals())

ReifyArguments = Instruction("ReifyArguments", 118, [Reg8], globals())

CreateRegExp = Instruction(
    "CreateRegExp", 119, [Reg8, UInt32, UInt32, UInt32], globals()
)

CreateRegExp.operands[1].operand_meaning = OperandMeaning.string_id

CreateRegExp.operands[2].operand_meaning = OperandMeaning.string_id

SwitchImm = Instruction(
    "SwitchImm", 120, [Reg8, UInt32, Addr32, UInt32, UInt32], globals()
)

StartGenerator = Instruction("StartGenerator", 121, [], globals())

ResumeGenerator = Instruction("ResumeGenerator", 122, [Reg8, Reg8], globals())

CompleteGenerator = Instruction("CompleteGenerator", 123, [], globals())

CreateGenerator = Instruction("CreateGenerator", 124, [Reg8, Reg8, UInt16], globals())

CreateGeneratorLongIndex = Instruction(
    "CreateGeneratorLongIndex", 125, [Reg8, Reg8, UInt32], globals()
)

Jmp = Instruction("Jmp", 126, [Addr8], globals())
JmpLong = Instruction("JmpLong", 127, [Addr32], globals())

JmpTrue = Instruction("JmpTrue", 128, [Addr8, Reg8], globals())
JmpTrueLong = Instruction("JmpTrueLong", 129, [Addr32, Reg8], globals())

JmpFalse = Instruction("JmpFalse", 130, [Addr8, Reg8], globals())
JmpFalseLong = Instruction("JmpFalseLong", 131, [Addr32, Reg8], globals())

JmpUndefined = Instruction("JmpUndefined", 132, [Addr8, Reg8], globals())
JmpUndefinedLong = Instruction("JmpUndefinedLong", 133, [Addr32, Reg8], globals())

SaveGenerator = Instruction("SaveGenerator", 134, [Addr8], globals())
SaveGeneratorLong = Instruction("SaveGeneratorLong", 135, [Addr32], globals())

JLess = Instruction("JLess", 136, [Addr8, Reg8, Reg8], globals())
JLessLong = Instruction("JLessLong", 137, [Addr32, Reg8, Reg8], globals())

JNotLess = Instruction("JNotLess", 138, [Addr8, Reg8, Reg8], globals())
JNotLessLong = Instruction("JNotLessLong", 139, [Addr32, Reg8, Reg8], globals())

JLessN = Instruction("JLessN", 140, [Addr8, Reg8, Reg8], globals())
JLessNLong = Instruction("JLessNLong", 141, [Addr32, Reg8, Reg8], globals())

JNotLessN = Instruction("JNotLessN", 142, [Addr8, Reg8, Reg8], globals())
JNotLessNLong = Instruction("JNotLessNLong", 143, [Addr32, Reg8, Reg8], globals())

JLessEqual = Instruction("JLessEqual", 144, [Addr8, Reg8, Reg8], globals())
JLessEqualLong = Instruction("JLessEqualLong", 145, [Addr32, Reg8, Reg8], globals())

JNotLessEqual = Instruction("JNotLessEqual", 146, [Addr8, Reg8, Reg8], globals())
JNotLessEqualLong = Instruction(
    "JNotLessEqualLong", 147, [Addr32, Reg8, Reg8], globals()
)

JLessEqualN = Instruction("JLessEqualN", 148, [Addr8, Reg8, Reg8], globals())
JLessEqualNLong = Instruction("JLessEqualNLong", 149, [Addr32, Reg8, Reg8], globals())

JNotLessEqualN = Instruction("JNotLessEqualN", 150, [Addr8, Reg8, Reg8], globals())
JNotLessEqualNLong = Instruction(
    "JNotLessEqualNLong", 151, [Addr32, Reg8, Reg8], globals()
)

JGreater = Instruction("JGreater", 152, [Addr8, Reg8, Reg8], globals())
JGreaterLong = Instruction("JGreaterLong", 153, [Addr32, Reg8, Reg8], globals())

JNotGreater = Instruction("JNotGreater", 154, [Addr8, Reg8, Reg8], globals())
JNotGreaterLong = Instruction("JNotGreaterLong", 155, [Addr32, Reg8, Reg8], globals())

JGreaterN = Instruction("JGreaterN", 156, [Addr8, Reg8, Reg8], globals())
JGreaterNLong = Instruction("JGreaterNLong", 157, [Addr32, Reg8, Reg8], globals())

JNotGreaterN = Instruction("JNotGreaterN", 158, [Addr8, Reg8, Reg8], globals())
JNotGreaterNLong = Instruction("JNotGreaterNLong", 159, [Addr32, Reg8, Reg8], globals())

JGreaterEqual = Instruction("JGreaterEqual", 160, [Addr8, Reg8, Reg8], globals())
JGreaterEqualLong = Instruction(
    "JGreaterEqualLong", 161, [Addr32, Reg8, Reg8], globals()
)

JNotGreaterEqual = Instruction("JNotGreaterEqual", 162, [Addr8, Reg8, Reg8], globals())
JNotGreaterEqualLong = Instruction(
    "JNotGreaterEqualLong", 163, [Addr32, Reg8, Reg8], globals()
)

JGreaterEqualN = Instruction("JGreaterEqualN", 164, [Addr8, Reg8, Reg8], globals())
JGreaterEqualNLong = Instruction(
    "JGreaterEqualNLong", 165, [Addr32, Reg8, Reg8], globals()
)

JNotGreaterEqualN = Instruction(
    "JNotGreaterEqualN", 166, [Addr8, Reg8, Reg8], globals()
)
JNotGreaterEqualNLong = Instruction(
    "JNotGreaterEqualNLong", 167, [Addr32, Reg8, Reg8], globals()
)

JEqual = Instruction("JEqual", 168, [Addr8, Reg8, Reg8], globals())
JEqualLong = Instruction("JEqualLong", 169, [Addr32, Reg8, Reg8], globals())

JNotEqual = Instruction("JNotEqual", 170, [Addr8, Reg8, Reg8], globals())
JNotEqualLong = Instruction("JNotEqualLong", 171, [Addr32, Reg8, Reg8], globals())

JStrictEqual = Instruction("JStrictEqual", 172, [Addr8, Reg8, Reg8], globals())
JStrictEqualLong = Instruction("JStrictEqualLong", 173, [Addr32, Reg8, Reg8], globals())

JStrictNotEqual = Instruction("JStrictNotEqual", 174, [Addr8, Reg8, Reg8], globals())
JStrictNotEqualLong = Instruction(
    "JStrictNotEqualLong", 175, [Addr32, Reg8, Reg8], globals()
)

CallDirect.operands[2].operand_meaning = OperandMeaning.function_id

CreateClosure.operands[2].operand_meaning = OperandMeaning.function_id

CreateClosureLongIndex.operands[2].operand_meaning = OperandMeaning.function_id

CreateGeneratorClosure.operands[2].operand_meaning = OperandMeaning.function_id

CreateGeneratorClosureLongIndex.operands[2].operand_meaning = OperandMeaning.function_id

CreateGenerator.operands[2].operand_meaning = OperandMeaning.function_id

CreateGeneratorLongIndex.operands[2].operand_meaning = OperandMeaning.function_id

_opcode_to_instruction: Dict[int, Instruction] = {v.opcode: v for v in _instructions}
_name_to_instruction: Dict[str, Instruction] = {v.name: v for v in _instructions}

_builtin_function_names: List[str] = [
    "Array.isArray",
    "ArrayBuffer.isView",
    "Date.UTC",
    "Date.now",
    "Date.parse",
    "HermesInternal.getEpilogues",
    "HermesInternal.silentSetPrototypeOf",
    "HermesInternal.requireFast",
    "HermesInternal.getTemplateObject",
    "HermesInternal.ensureObject",
    "HermesInternal.copyDataProperties",
    "HermesInternal.copyRestArgs",
    "HermesInternal.exportAll",
    "JSON.parse",
    "JSON.stringify",
    "Math.abs",
    "Math.acos",
    "Math.asin",
    "Math.atan",
    "Math.atan2",
    "Math.ceil",
    "Math.cos",
    "Math.exp",
    "Math.floor",
    "Math.hypot",
    "Math.imul",
    "Math.log",
    "Math.max",
    "Math.min",
    "Math.pow",
    "Math.random",
    "Math.round",
    "Math.sin",
    "Math.sqrt",
    "Math.tan",
    "Math.trunc",
    "Object.create",
    "Object.defineProperties",
    "Object.defineProperty",
    "Object.freeze",
    "Object.getOwnPropertyDescriptor",
    "Object.getOwnPropertyNames",
    "Object.getPrototypeOf",
    "Object.isExtensible",
    "Object.isFrozen",
    "Object.keys",
    "Object.seal",
    "String.fromCharCode",
]



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from typing import List, Tuple, Dict, Set, Sequence, Union, Optional, Any
from os.path import dirname, realpath
from collections import defaultdict

from defs import HermesDecompiler, BasicBlock, DecompiledFunctionBody
from hbc_bytecode_parser import parse_hbc_bytecode


def pass1_set_metadata(state: HermesDecompiler, function_body: DecompiledFunctionBody):

    function_body.function_name = state.hbc_reader.strings[
        function_body.function_object.functionName
    ]
    function_body.nested_frames = []
    function_body.ret_anchors = {}
    function_body.throw_anchors = {}
    function_body.jump_anchors = {}
    function_body.jump_targets = set()

    # Record the addresses for the try_*, catch_* addresses within code

    function_body.try_starts = defaultdict(list)
    function_body.try_ends = defaultdict(list)
    function_body.catch_targets = defaultdict(list)
    if function_body.function_object.hasExceptionHandler:
        for handler_count, handler in enumerate(function_body.exc_handlers):
            function_body.try_starts[handler.start].append(
                "try_start_%d" % handler_count
            )
            function_body.try_ends[handler.end].append("try_end%d" % handler_count)
            function_body.catch_targets[handler.target].append(
                "catch_target%d" % handler_count
            )

    # As well as Jump, Switch, Yield

    for instruction in parse_hbc_bytecode(
        function_body.function_object, state.hbc_reader
    ):
        if instruction.inst.name[0] == "J" or instruction.inst.name.startswith(
            "SaveGenerator"
        ):
            function_body.jump_anchors[instruction.next_pos] = instruction
            function_body.jump_targets.add(instruction.original_pos + instruction.arg1)

        elif instruction.inst.name == "SwitchImm":
            function_body.jump_anchors[instruction.next_pos] = instruction
            function_body.jump_targets.add(instruction.original_pos + instruction.arg3)
            for jump_target in instruction.switch_jump_table:
                function_body.jump_targets.add(jump_target)

        elif instruction.inst.name == "Ret":
            function_body.ret_anchors[instruction.next_pos] = instruction

        elif instruction.inst.name == "Throw":
            function_body.throw_anchors[instruction.next_pos] = instruction

    """
        Make basic blocks out of assembly
    """

    error_handlers = (
        state.hbc_reader.function_id_to_exc_handlers[function_body.function_id]
        if function_body.function_object.hasExceptionHandler
        else []
    )

    basic_block_boundaries = (
        {0}
        | function_body.try_starts.keys()
        | function_body.try_ends.keys()
        | function_body.catch_targets.keys()
        | function_body.jump_anchors.keys()
        | function_body.ret_anchors.keys()
        | function_body.throw_anchors.keys()
        | function_body.jump_targets
    )

    basic_blocks = function_body.basic_blocks = []

    start_to_basic_block: Dict[int, BasicBlock] = {}
    end_to_basic_block: Dict[int, BasicBlock] = {}

    # Create the basic blocks that will constitute the control
    # flow graph for the current function

    basic_block_start = 0
    may_have_fallen_through: bool = False
    for basic_block_end in sorted(basic_block_boundaries)[1:]:
        basic_block = BasicBlock()
        basic_block.start_address = basic_block_start
        basic_block.end_address = basic_block_end

        start_to_basic_block[basic_block_start] = basic_block
        end_to_basic_block[basic_block_end] = basic_block

        basic_block.child_nodes = []
        basic_block.parent_nodes = []
        basic_block.error_handling_child_nodes = []
        basic_block.error_handling_parent_nodes = []

        if may_have_fallen_through:
            basic_block.parent_nodes.append(basic_blocks[-1])
            basic_blocks[-1].child_nodes.append(basic_block)

        may_have_fallen_through = True
        if basic_block_end in function_body.ret_anchors:
            may_have_fallen_through = False
            basic_block.anchor_instruction = function_body.ret_anchors[basic_block_end]
            basic_block.is_unconditional_return_end = True
        elif basic_block_end in function_body.throw_anchors:
            may_have_fallen_through = False
            basic_block.anchor_instruction = function_body.throw_anchors[
                basic_block_end
            ]
            basic_block.is_unconditional_throw_anchor = True
        elif basic_block_end in function_body.jump_anchors:
            op = function_body.jump_anchors[basic_block_end]
            basic_block.anchor_instruction = op
            op_name = op.inst.name
            if op_name in ("Jmp", "JmpLong"):
                may_have_fallen_through = False
                basic_block.jump_targets_for_anchor = [op.original_pos + op.arg1]
                basic_block.is_unconditional_jump_anchor = True
            elif op_name == "SwitchImm":
                may_have_fallen_through = False
                basic_block.jump_targets_for_anchor = sorted(
                    {op.original_pos + op.arg3, *op.switch_jump_table}
                )
                basic_block.if_switch_action_anchor = True
            elif op_name in ("SaveGenerator", "SaveGeneratorLong"):
                may_have_fallen_through = True
                basic_block.jump_targets_for_anchor = [op.original_pos + op.arg1]
                basic_block.is_yield_action_anchor = True
            elif op_name[0] == "J":
                may_have_fallen_through = True
                basic_block.jump_targets_for_anchor = [op.original_pos + op.arg1]
                basic_block.is_conditional_jump_anchor = True
            else:
                raise ValueError

        basic_blocks.append(basic_block)

        basic_block_start = basic_block_end

    for basic_block in basic_blocks:
        if basic_block.jump_targets_for_anchor:
            # Link graphes between these in case of
            # jump/switch/yield instructions

            for jump_target in basic_block.jump_targets_for_anchor:
                jump_target_block = start_to_basic_block[jump_target]
                if jump_target_block not in basic_block.child_nodes:
                    basic_block.child_nodes.append(jump_target_block)
                if basic_block not in jump_target_block.parent_nodes:
                    jump_target_block.parent_nodes.append(basic_block)

        # Link graphes between these when error
        # handling is present

        for error_handler in error_handlers:
            if (
                basic_block.start_address
                <= error_handler.start
                < basic_block.end_address
            ) or (
                basic_block.start_address < error_handler.end <= basic_block.end_address
            ):
                error_handler_block = start_to_basic_block[error_handler.target]

                if error_handler_block not in basic_block.error_handling_child_nodes:
                    basic_block.error_handling_child_nodes.append(error_handler_block)
                if basic_block not in error_handler_block.error_handling_parent_nodes:
                    error_handler_block.error_handling_parent_nodes.append(basic_block)

    # WIP...



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from os.path import realpath, dirname
from argparse import ArgumentParser
import sys

SCRIPT_DIR = dirname(realpath(__file__))
ROOT_DIR = realpath(SCRIPT_DIR + "/..")
PARSERS_DIR = realpath(ROOT_DIR + "/parsers")

sys.path.insert(0, SCRIPT_DIR)
sys.path.insert(0, PARSERS_DIR)

from hbc_file_parser import HBCReader
from pass1_set_metadata import pass1_set_metadata
from pass2_transform_code import pass2_transform_code
from pass3_parse_forin_loops import pass3_parse_forin_loops
from pass4_name_closure_vars import pass4_name_closure_vars
from defs import HermesDecompiler, FunctionTableIndex, DecompiledFunctionBody

"""
    Entry points for the Hermes HBC Decompiler
"""

# Decompile a function and its nested closures to the standard
# output, given a parsed Hermes bytecode file and a function ID:


def decompile_function(state: HermesDecompiler, function_id: int, **kwargs):

    dehydrated = DecompiledFunctionBody()

    dehydrated.function_id = function_id
    dehydrated.function_object = state.hbc_reader.function_headers[function_id]
    dehydrated.is_global = function_id == state.hbc_reader.header.globalCodeIndex

    # Used within FunctionTableIndex.closure_decompile to
    # provide extra context about the current function:
    for key, value in kwargs.items():
        setattr(dehydrated, key, value)

    if dehydrated.function_object.hasExceptionHandler:
        dehydrated.exc_handlers = state.hbc_reader.function_id_to_exc_handlers[
            function_id
        ]

    pass1_set_metadata(state, dehydrated)

    pass2_transform_code(state, dehydrated)

    pass3_parse_forin_loops(state, dehydrated)

    pass4_name_closure_vars(state, dehydrated)

    dehydrated.output_code(state)


def do_decompilation(state: HermesDecompiler, file_handle):

    hbc_reader = HBCReader()
    state.hbc_reader = hbc_reader

    state.hbc_reader.read_whole_file(file_handle)

    state.calldirect_function_ids = set()

    global_function_index = state.hbc_reader.header.globalCodeIndex

    state.indent_level = 0
    decompile_function(state, global_function_index)
    print()

    for function_id in sorted(state.calldirect_function_ids):
        decompile_function(state, function_id)
        print()


def main():

    args = ArgumentParser()

    args.add_argument("input_file")
    args.add_argument("output_file", nargs="?")

    args = args.parse_args()

    state = HermesDecompiler()
    state.input_file = args.input_file
    state.output_file = args.output_file

    with open(state.input_file, "rb") as file_handle:
        if state.output_file:
            stdout = sys.stdout
            with open(state.output_file, "w", encoding="utf-8") as sys.stdout:
                do_decompilation(state, file_handle)
            sys.stdout = stdout

            print()
            print('[+] Decompiled output wrote to "%s"' % state.output_file)
            print()
        else:
            sys.stdout.reconfigure(encoding="utf-8")
            do_decompilation(state, file_handle)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from typing import List, Tuple, Dict, Set, Sequence, Union, Optional, Any
from os.path import dirname, realpath
from dataclasses import dataclass

from defs import (
    HermesDecompiler,
    DecompiledFunctionBody,
    NestedFrame,
    Environment,
    TokenString,
    ForInLoopInit,
    ForInLoopNextIter,
    RawToken,
    ReturnDirective,
    ThrowDirective,
    JumpNotCondition,
    JumpCondition,
    AssignmentToken,
    BasicBlock,
    LeftParenthesisToken,
    RightHandRegToken,
    RightParenthesisToken,
    LeftHandRegToken,
    NewEnvironmentToken,
    StoreToEnvironment,
    GetEnvironmentToken,
    LoadFromEnvironmentToken,
    FunctionTableIndex,
)


def pass3_parse_forin_loops(
    state: HermesDecompiler, function_body: DecompiledFunctionBody
):

    JNC = JumpNotCondition  # Forward jumps
    JC = JumpCondition  # Backwards jumps

    RD = ReturnDirective
    TD = ThrowDirective

    AT = AssignmentToken

    TS = TokenString
    RT = RawToken
    LHRT = LeftHandRegToken
    RHRT = RightHandRegToken

    FILI = ForInLoopInit
    FILNI = ForInLoopNextIter

    LPT = LeftParenthesisToken
    RPT = RightParenthesisToken

    # Implementation for recreating "for..in" structures (see the .odt notes)

    lines = function_body.statements

    for index, line in enumerate(lines):
        # Seek for a GetPNameList instruction
        has_fili = line.tokens and isinstance(line.tokens[0], FILI)

        if has_fili:
            # Seek for a GetNextPName instruction further it
            other_index = next(
                (
                    other_index
                    for other_index in range(index, len(lines))
                    if isinstance(lines[other_index].tokens[0], FILNI)
                ),
                None,
            )
            assert other_index

            # The beginning point of the loop's inner basic block
            # will be:
            begin_address = lines[index + 2].assembly[0].original_pos

            if isinstance(lines[index + 1].tokens[0], JC) and isinstance(
                lines[other_index + 1].tokens[0], JC
            ):
                # There is a nested loop and we're going to an
                # instruction located backwards in the HBC after
                # the end of the current for..in loop (hence the JC
                # rather than JNC token in our intermediate
                # representation), manage to find the end of the
                # current basic block through iterating over code

                continue  # Commented yet: Show not to always work

                """
                for further_index in range(other_index + 2, len(lines)):
                    
                    if (isinstance(lines[further_index].tokens[0], JC) and
                        lines[further_index].tokens[0].target_address == begin_address and
                        lines[further_index].tokens[1] == RT('true')):
                        
                        # We've found an unconditional backwards jump towards
                        # the beginning of the loop, close the current basic
                        # block
                        break
                    
                    if (isinstance(lines[further_index].tokens[0], RD) or
                        isinstance(lines[further_index].tokens[0], TD)):
                        
                        # We've found either a "return" or a "throw"
                        # instruction, likely located after a conditional
                        # jump towards the beginning of the current basic
                        # block, close the current basic block
                        break
                
                end_address = lines[further_index + 1].assembly[0].original_pos
                """

            else:
                # There is no nested loop and we're having the end
                # address of the current loop located directly in
                # the first conditional jump located after the
                # respective GetPNameList and GetNextPName
                # instructions within the original HBC

                end_address = lines[other_index + 1].tokens[0].target_address

                assert isinstance(lines[index + 1].tokens[0], JNC)
                assert isinstance(lines[other_index + 1].tokens[0], JNC)

                boundary_1 = lines[index + 1].tokens[0].target_address
                boundary_2 = lines[other_index + 1].tokens[0].target_address

                min_boundary = min(boundary_1, boundary_2)
                max_boundary = max(boundary_1, boundary_2)
                if min_boundary != max_boundary:
                    weird_case = False
                    for further_index in range(other_index + 2, len(lines)):
                        if (
                            min_boundary
                            <= lines[further_index].assembly[0].original_pos
                            < max_boundary
                        ):
                            weird_case = True
                            # lines[further_index].tokens.insert(0, RT('// NOTE: Orphan basic block present beween two condition branches automatically created upon For..in loop: '))
                            break
                    if weird_case:
                        continue

            function_body.nested_frames.append(NestedFrame(begin_address, end_address))

            line.tokens = [
                RT("for"),
                LPT(),
                LHRT(lines[other_index].tokens[0].next_value_register),
                RT(" in "),
                RHRT(line.tokens[0].obj_register),
                RPT(),
            ]
            lines[index + 1].tokens = []
            lines[other_index].tokens = []
            lines[other_index + 1].tokens = []

            continue



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
"""
Note: The contents of the current file have been automatically
generated by the "utils/hermes_bytecode_structs_parser.py"
script

Please do not edit it manually. 👍
"""

from typing import List, Set, Dict, Union, Optional, Sequence, Any

# Imports from the current diretory
from .def_classes import *

_instructions: List[Instruction] = []

Reg8 = OperandType("Reg8", "uint8_t")
Reg32 = OperandType("Reg32", "uint32_t")
UInt8 = OperandType("UInt8", "uint8_t")
UInt16 = OperandType("UInt16", "uint16_t")
UInt32 = OperandType("UInt32", "uint32_t")
Addr8 = OperandType("Addr8", "int8_t")
Addr32 = OperandType("Addr32", "int32_t")
Imm32 = OperandType("Imm32", "int32_t")
Double = OperandType("Double", "double")

Unreachable = Instruction("Unreachable", 0, [], globals())

NewObjectWithBuffer = Instruction(
    "NewObjectWithBuffer", 1, [Reg8, UInt16, UInt16, UInt16, UInt16], globals()
)

NewObjectWithBufferLong = Instruction(
    "NewObjectWithBufferLong", 2, [Reg8, UInt16, UInt16, UInt32, UInt32], globals()
)

NewObject = Instruction("NewObject", 3, [Reg8], globals())

NewObjectWithParent = Instruction("NewObjectWithParent", 4, [Reg8, Reg8], globals())

NewArrayWithBuffer = Instruction(
    "NewArrayWithBuffer", 5, [Reg8, UInt16, UInt16, UInt16], globals()
)

NewArrayWithBufferLong = Instruction(
    "NewArrayWithBufferLong", 6, [Reg8, UInt16, UInt16, UInt32], globals()
)

NewArray = Instruction("NewArray", 7, [Reg8, UInt16], globals())

Mov = Instruction("Mov", 8, [Reg8, Reg8], globals())

MovLong = Instruction("MovLong", 9, [Reg32, Reg32], globals())

Negate = Instruction("Negate", 10, [Reg8, Reg8], globals())

Not = Instruction("Not", 11, [Reg8, Reg8], globals())

BitNot = Instruction("BitNot", 12, [Reg8, Reg8], globals())

TypeOf = Instruction("TypeOf", 13, [Reg8, Reg8], globals())

Eq = Instruction("Eq", 14, [Reg8, Reg8, Reg8], globals())

StrictEq = Instruction("StrictEq", 15, [Reg8, Reg8, Reg8], globals())

Neq = Instruction("Neq", 16, [Reg8, Reg8, Reg8], globals())

StrictNeq = Instruction("StrictNeq", 17, [Reg8, Reg8, Reg8], globals())

Less = Instruction("Less", 18, [Reg8, Reg8, Reg8], globals())

LessEq = Instruction("LessEq", 19, [Reg8, Reg8, Reg8], globals())

Greater = Instruction("Greater", 20, [Reg8, Reg8, Reg8], globals())

GreaterEq = Instruction("GreaterEq", 21, [Reg8, Reg8, Reg8], globals())

Add = Instruction("Add", 22, [Reg8, Reg8, Reg8], globals())

AddN = Instruction("AddN", 23, [Reg8, Reg8, Reg8], globals())

Mul = Instruction("Mul", 24, [Reg8, Reg8, Reg8], globals())

MulN = Instruction("MulN", 25, [Reg8, Reg8, Reg8], globals())

Div = Instruction("Div", 26, [Reg8, Reg8, Reg8], globals())

DivN = Instruction("DivN", 27, [Reg8, Reg8, Reg8], globals())

Mod = Instruction("Mod", 28, [Reg8, Reg8, Reg8], globals())

Sub = Instruction("Sub", 29, [Reg8, Reg8, Reg8], globals())

SubN = Instruction("SubN", 30, [Reg8, Reg8, Reg8], globals())

LShift = Instruction("LShift", 31, [Reg8, Reg8, Reg8], globals())

RShift = Instruction("RShift", 32, [Reg8, Reg8, Reg8], globals())

URshift = Instruction("URshift", 33, [Reg8, Reg8, Reg8], globals())

BitAnd = Instruction("BitAnd", 34, [Reg8, Reg8, Reg8], globals())

BitXor = Instruction("BitXor", 35, [Reg8, Reg8, Reg8], globals())

BitOr = Instruction("BitOr", 36, [Reg8, Reg8, Reg8], globals())

Inc = Instruction("Inc", 37, [Reg8, Reg8], globals())

Dec = Instruction("Dec", 38, [Reg8, Reg8], globals())

InstanceOf = Instruction("InstanceOf", 39, [Reg8, Reg8, Reg8], globals())

IsIn = Instruction("IsIn", 40, [Reg8, Reg8, Reg8], globals())

GetEnvironment = Instruction("GetEnvironment", 41, [Reg8, UInt8], globals())

StoreToEnvironment = Instruction(
    "StoreToEnvironment", 42, [Reg8, UInt8, Reg8], globals()
)

StoreToEnvironmentL = Instruction(
    "StoreToEnvironmentL", 43, [Reg8, UInt16, Reg8], globals()
)

StoreNPToEnvironment = Instruction(
    "StoreNPToEnvironment", 44, [Reg8, UInt8, Reg8], globals()
)

StoreNPToEnvironmentL = Instruction(
    "StoreNPToEnvironmentL", 45, [Reg8, UInt16, Reg8], globals()
)

LoadFromEnvironment = Instruction(
    "LoadFromEnvironment", 46, [Reg8, Reg8, UInt8], globals()
)

LoadFromEnvironmentL = Instruction(
    "LoadFromEnvironmentL", 47, [Reg8, Reg8, UInt16], globals()
)

GetGlobalObject = Instruction("GetGlobalObject", 48, [Reg8], globals())

GetNewTarget = Instruction("GetNewTarget", 49, [Reg8], globals())

CreateEnvironment = Instruction("CreateEnvironment", 50, [Reg8], globals())

CreateInnerEnvironment = Instruction(
    "CreateInnerEnvironment", 51, [Reg8, Reg8, UInt32], globals()
)

DeclareGlobalVar = Instruction("DeclareGlobalVar", 52, [UInt32], globals())

DeclareGlobalVar.operands[0].operand_meaning = OperandMeaning.string_id

ThrowIfHasRestrictedGlobalProperty = Instruction(
    "ThrowIfHasRestrictedGlobalProperty", 53, [UInt32], globals()
)

ThrowIfHasRestrictedGlobalProperty.operands[
    0
].operand_meaning = OperandMeaning.string_id

GetByIdShort = Instruction("GetByIdShort", 54, [Reg8, Reg8, UInt8, UInt8], globals())

GetById = Instruction("GetById", 55, [Reg8, Reg8, UInt8, UInt16], globals())

GetByIdLong = Instruction("GetByIdLong", 56, [Reg8, Reg8, UInt8, UInt32], globals())

GetByIdShort.operands[3].operand_meaning = OperandMeaning.string_id

GetById.operands[3].operand_meaning = OperandMeaning.string_id

GetByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

TryGetById = Instruction("TryGetById", 57, [Reg8, Reg8, UInt8, UInt16], globals())

TryGetByIdLong = Instruction(
    "TryGetByIdLong", 58, [Reg8, Reg8, UInt8, UInt32], globals()
)

TryGetById.operands[3].operand_meaning = OperandMeaning.string_id

TryGetByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

PutById = Instruction("PutById", 59, [Reg8, Reg8, UInt8, UInt16], globals())

PutByIdLong = Instruction("PutByIdLong", 60, [Reg8, Reg8, UInt8, UInt32], globals())

PutById.operands[3].operand_meaning = OperandMeaning.string_id

PutByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

TryPutById = Instruction("TryPutById", 61, [Reg8, Reg8, UInt8, UInt16], globals())

TryPutByIdLong = Instruction(
    "TryPutByIdLong", 62, [Reg8, Reg8, UInt8, UInt32], globals()
)

TryPutById.operands[3].operand_meaning = OperandMeaning.string_id

TryPutByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

PutNewOwnByIdShort = Instruction(
    "PutNewOwnByIdShort", 63, [Reg8, Reg8, UInt8], globals()
)

PutNewOwnById = Instruction("PutNewOwnById", 64, [Reg8, Reg8, UInt16], globals())

PutNewOwnByIdLong = Instruction(
    "PutNewOwnByIdLong", 65, [Reg8, Reg8, UInt32], globals()
)

PutNewOwnByIdShort.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnById.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnByIdLong.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnNEById = Instruction("PutNewOwnNEById", 66, [Reg8, Reg8, UInt16], globals())

PutNewOwnNEByIdLong = Instruction(
    "PutNewOwnNEByIdLong", 67, [Reg8, Reg8, UInt32], globals()
)

PutNewOwnNEById.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnNEByIdLong.operands[2].operand_meaning = OperandMeaning.string_id

PutOwnByIndex = Instruction("PutOwnByIndex", 68, [Reg8, Reg8, UInt8], globals())

PutOwnByIndexL = Instruction("PutOwnByIndexL", 69, [Reg8, Reg8, UInt32], globals())

PutOwnByVal = Instruction("PutOwnByVal", 70, [Reg8, Reg8, Reg8, UInt8], globals())

DelById = Instruction("DelById", 71, [Reg8, Reg8, UInt16], globals())

DelByIdLong = Instruction("DelByIdLong", 72, [Reg8, Reg8, UInt32], globals())

DelById.operands[2].operand_meaning = OperandMeaning.string_id

DelByIdLong.operands[2].operand_meaning = OperandMeaning.string_id

GetByVal = Instruction("GetByVal", 73, [Reg8, Reg8, Reg8], globals())

PutByVal = Instruction("PutByVal", 74, [Reg8, Reg8, Reg8], globals())

DelByVal = Instruction("DelByVal", 75, [Reg8, Reg8, Reg8], globals())

PutOwnGetterSetterByVal = Instruction(
    "PutOwnGetterSetterByVal", 76, [Reg8, Reg8, Reg8, Reg8, UInt8], globals()
)

GetPNameList = Instruction("GetPNameList", 77, [Reg8, Reg8, Reg8, Reg8], globals())

GetNextPName = Instruction(
    "GetNextPName", 78, [Reg8, Reg8, Reg8, Reg8, Reg8], globals()
)

Call = Instruction("Call", 79, [Reg8, Reg8, UInt8], globals())
Call.has_ret_target = True

Construct = Instruction("Construct", 80, [Reg8, Reg8, UInt8], globals())
Construct.has_ret_target = True

Call1 = Instruction("Call1", 81, [Reg8, Reg8, Reg8], globals())
Call1.has_ret_target = True

CallDirect = Instruction("CallDirect", 82, [Reg8, UInt8, UInt16], globals())

CallDirect.operands[2].operand_meaning = OperandMeaning.function_id
CallDirect.has_ret_target = True

Call2 = Instruction("Call2", 83, [Reg8, Reg8, Reg8, Reg8], globals())
Call2.has_ret_target = True

Call3 = Instruction("Call3", 84, [Reg8, Reg8, Reg8, Reg8, Reg8], globals())
Call3.has_ret_target = True

Call4 = Instruction("Call4", 85, [Reg8, Reg8, Reg8, Reg8, Reg8, Reg8], globals())
Call4.has_ret_target = True

CallLong = Instruction("CallLong", 86, [Reg8, Reg8, UInt32], globals())
CallLong.has_ret_target = True

ConstructLong = Instruction("ConstructLong", 87, [Reg8, Reg8, UInt32], globals())
ConstructLong.has_ret_target = True

CallDirectLongIndex = Instruction(
    "CallDirectLongIndex", 88, [Reg8, UInt8, UInt32], globals()
)
CallDirectLongIndex.has_ret_target = True

CallBuiltin = Instruction("CallBuiltin", 89, [Reg8, UInt8, UInt8], globals())

CallBuiltinLong = Instruction("CallBuiltinLong", 90, [Reg8, UInt8, UInt32], globals())

GetBuiltinClosure = Instruction("GetBuiltinClosure", 91, [Reg8, UInt8], globals())

Ret = Instruction("Ret", 92, [Reg8], globals())

Catch = Instruction("Catch", 93, [Reg8], globals())

DirectEval = Instruction("DirectEval", 94, [Reg8, Reg8], globals())

Throw = Instruction("Throw", 95, [Reg8], globals())

ThrowIfEmpty = Instruction("ThrowIfEmpty", 96, [Reg8, Reg8], globals())

Debugger = Instruction("Debugger", 97, [], globals())

AsyncBreakCheck = Instruction("AsyncBreakCheck", 98, [], globals())

ProfilePoint = Instruction("ProfilePoint", 99, [UInt16], globals())

CreateClosure = Instruction("CreateClosure", 100, [Reg8, Reg8, UInt16], globals())

CreateClosureLongIndex = Instruction(
    "CreateClosureLongIndex", 101, [Reg8, Reg8, UInt32], globals()
)

CreateClosure.operands[2].operand_meaning = OperandMeaning.function_id

CreateClosureLongIndex.operands[2].operand_meaning = OperandMeaning.function_id

CreateGeneratorClosure = Instruction(
    "CreateGeneratorClosure", 102, [Reg8, Reg8, UInt16], globals()
)

CreateGeneratorClosureLongIndex = Instruction(
    "CreateGeneratorClosureLongIndex", 103, [Reg8, Reg8, UInt32], globals()
)

CreateGeneratorClosure.operands[2].operand_meaning = OperandMeaning.function_id

CreateGeneratorClosureLongIndex.operands[2].operand_meaning = OperandMeaning.function_id

CreateAsyncClosure = Instruction(
    "CreateAsyncClosure", 104, [Reg8, Reg8, UInt16], globals()
)

CreateAsyncClosureLongIndex = Instruction(
    "CreateAsyncClosureLongIndex", 105, [Reg8, Reg8, UInt32], globals()
)

CreateAsyncClosure.operands[2].operand_meaning = OperandMeaning.function_id

CreateAsyncClosureLongIndex.operands[2].operand_meaning = OperandMeaning.function_id

CreateThis = Instruction("CreateThis", 106, [Reg8, Reg8, Reg8], globals())

SelectObject = Instruction("SelectObject", 107, [Reg8, Reg8, Reg8], globals())

LoadParam = Instruction("LoadParam", 108, [Reg8, UInt8], globals())

LoadParamLong = Instruction("LoadParamLong", 109, [Reg8, UInt32], globals())

LoadConstUInt8 = Instruction("LoadConstUInt8", 110, [Reg8, UInt8], globals())

LoadConstInt = Instruction("LoadConstInt", 111, [Reg8, Imm32], globals())

LoadConstDouble = Instruction("LoadConstDouble", 112, [Reg8, Double], globals())

LoadConstBigInt = Instruction("LoadConstBigInt", 113, [Reg8, UInt16], globals())

LoadConstBigIntLongIndex = Instruction(
    "LoadConstBigIntLongIndex", 114, [Reg8, UInt32], globals()
)

LoadConstBigInt.operands[1].operand_meaning = OperandMeaning.bigint_id

LoadConstBigIntLongIndex.operands[1].operand_meaning = OperandMeaning.bigint_id

LoadConstString = Instruction("LoadConstString", 115, [Reg8, UInt16], globals())

LoadConstStringLongIndex = Instruction(
    "LoadConstStringLongIndex", 116, [Reg8, UInt32], globals()
)

LoadConstString.operands[1].operand_meaning = OperandMeaning.string_id

LoadConstStringLongIndex.operands[1].operand_meaning = OperandMeaning.string_id

LoadConstEmpty = Instruction("LoadConstEmpty", 117, [Reg8], globals())

LoadConstUndefined = Instruction("LoadConstUndefined", 118, [Reg8], globals())

LoadConstNull = Instruction("LoadConstNull", 119, [Reg8], globals())

LoadConstTrue = Instruction("LoadConstTrue", 120, [Reg8], globals())

LoadConstFalse = Instruction("LoadConstFalse", 121, [Reg8], globals())

LoadConstZero = Instruction("LoadConstZero", 122, [Reg8], globals())

CoerceThisNS = Instruction("CoerceThisNS", 123, [Reg8, Reg8], globals())

LoadThisNS = Instruction("LoadThisNS", 124, [Reg8], globals())

ToNumber = Instruction("ToNumber", 125, [Reg8, Reg8], globals())

ToNumeric = Instruction("ToNumeric", 126, [Reg8, Reg8], globals())

ToInt32 = Instruction("ToInt32", 127, [Reg8, Reg8], globals())

AddEmptyString = Instruction("AddEmptyString", 128, [Reg8, Reg8], globals())

GetArgumentsPropByVal = Instruction(
    "GetArgumentsPropByVal", 129, [Reg8, Reg8, Reg8], globals()
)

GetArgumentsLength = Instruction("GetArgumentsLength", 130, [Reg8, Reg8], globals())

ReifyArguments = Instruction("ReifyArguments", 131, [Reg8], globals())

CreateRegExp = Instruction(
    "CreateRegExp", 132, [Reg8, UInt32, UInt32, UInt32], globals()
)

CreateRegExp.operands[1].operand_meaning = OperandMeaning.string_id

CreateRegExp.operands[2].operand_meaning = OperandMeaning.string_id

SwitchImm = Instruction(
    "SwitchImm", 133, [Reg8, UInt32, Addr32, UInt32, UInt32], globals()
)

StartGenerator = Instruction("StartGenerator", 134, [], globals())

ResumeGenerator = Instruction("ResumeGenerator", 135, [Reg8, Reg8], globals())

CompleteGenerator = Instruction("CompleteGenerator", 136, [], globals())

CreateGenerator = Instruction("CreateGenerator", 137, [Reg8, Reg8, UInt16], globals())

CreateGeneratorLongIndex = Instruction(
    "CreateGeneratorLongIndex", 138, [Reg8, Reg8, UInt32], globals()
)

CreateGenerator.operands[2].operand_meaning = OperandMeaning.function_id

CreateGeneratorLongIndex.operands[2].operand_meaning = OperandMeaning.function_id

IteratorBegin = Instruction("IteratorBegin", 139, [Reg8, Reg8], globals())

IteratorNext = Instruction("IteratorNext", 140, [Reg8, Reg8, Reg8], globals())

IteratorClose = Instruction("IteratorClose", 141, [Reg8, UInt8], globals())

Jmp = Instruction("Jmp", 142, [Addr8], globals())
JmpLong = Instruction("JmpLong", 143, [Addr32], globals())

JmpTrue = Instruction("JmpTrue", 144, [Addr8, Reg8], globals())
JmpTrueLong = Instruction("JmpTrueLong", 145, [Addr32, Reg8], globals())

JmpFalse = Instruction("JmpFalse", 146, [Addr8, Reg8], globals())
JmpFalseLong = Instruction("JmpFalseLong", 147, [Addr32, Reg8], globals())

JmpUndefined = Instruction("JmpUndefined", 148, [Addr8, Reg8], globals())
JmpUndefinedLong = Instruction("JmpUndefinedLong", 149, [Addr32, Reg8], globals())

SaveGenerator = Instruction("SaveGenerator", 150, [Addr8], globals())
SaveGeneratorLong = Instruction("SaveGeneratorLong", 151, [Addr32], globals())

JLess = Instruction("JLess", 152, [Addr8, Reg8, Reg8], globals())
JLessLong = Instruction("JLessLong", 153, [Addr32, Reg8, Reg8], globals())

JNotLess = Instruction("JNotLess", 154, [Addr8, Reg8, Reg8], globals())
JNotLessLong = Instruction("JNotLessLong", 155, [Addr32, Reg8, Reg8], globals())

JLessN = Instruction("JLessN", 156, [Addr8, Reg8, Reg8], globals())
JLessNLong = Instruction("JLessNLong", 157, [Addr32, Reg8, Reg8], globals())

JNotLessN = Instruction("JNotLessN", 158, [Addr8, Reg8, Reg8], globals())
JNotLessNLong = Instruction("JNotLessNLong", 159, [Addr32, Reg8, Reg8], globals())

JLessEqual = Instruction("JLessEqual", 160, [Addr8, Reg8, Reg8], globals())
JLessEqualLong = Instruction("JLessEqualLong", 161, [Addr32, Reg8, Reg8], globals())

JNotLessEqual = Instruction("JNotLessEqual", 162, [Addr8, Reg8, Reg8], globals())
JNotLessEqualLong = Instruction(
    "JNotLessEqualLong", 163, [Addr32, Reg8, Reg8], globals()
)

JLessEqualN = Instruction("JLessEqualN", 164, [Addr8, Reg8, Reg8], globals())
JLessEqualNLong = Instruction("JLessEqualNLong", 165, [Addr32, Reg8, Reg8], globals())

JNotLessEqualN = Instruction("JNotLessEqualN", 166, [Addr8, Reg8, Reg8], globals())
JNotLessEqualNLong = Instruction(
    "JNotLessEqualNLong", 167, [Addr32, Reg8, Reg8], globals()
)

JGreater = Instruction("JGreater", 168, [Addr8, Reg8, Reg8], globals())
JGreaterLong = Instruction("JGreaterLong", 169, [Addr32, Reg8, Reg8], globals())

JNotGreater = Instruction("JNotGreater", 170, [Addr8, Reg8, Reg8], globals())
JNotGreaterLong = Instruction("JNotGreaterLong", 171, [Addr32, Reg8, Reg8], globals())

JGreaterN = Instruction("JGreaterN", 172, [Addr8, Reg8, Reg8], globals())
JGreaterNLong = Instruction("JGreaterNLong", 173, [Addr32, Reg8, Reg8], globals())

JNotGreaterN = Instruction("JNotGreaterN", 174, [Addr8, Reg8, Reg8], globals())
JNotGreaterNLong = Instruction("JNotGreaterNLong", 175, [Addr32, Reg8, Reg8], globals())

JGreaterEqual = Instruction("JGreaterEqual", 176, [Addr8, Reg8, Reg8], globals())
JGreaterEqualLong = Instruction(
    "JGreaterEqualLong", 177, [Addr32, Reg8, Reg8], globals()
)

JNotGreaterEqual = Instruction("JNotGreaterEqual", 178, [Addr8, Reg8, Reg8], globals())
JNotGreaterEqualLong = Instruction(
    "JNotGreaterEqualLong", 179, [Addr32, Reg8, Reg8], globals()
)

JGreaterEqualN = Instruction("JGreaterEqualN", 180, [Addr8, Reg8, Reg8], globals())
JGreaterEqualNLong = Instruction(
    "JGreaterEqualNLong", 181, [Addr32, Reg8, Reg8], globals()
)

JNotGreaterEqualN = Instruction(
    "JNotGreaterEqualN", 182, [Addr8, Reg8, Reg8], globals()
)
JNotGreaterEqualNLong = Instruction(
    "JNotGreaterEqualNLong", 183, [Addr32, Reg8, Reg8], globals()
)

JEqual = Instruction("JEqual", 184, [Addr8, Reg8, Reg8], globals())
JEqualLong = Instruction("JEqualLong", 185, [Addr32, Reg8, Reg8], globals())

JNotEqual = Instruction("JNotEqual", 186, [Addr8, Reg8, Reg8], globals())
JNotEqualLong = Instruction("JNotEqualLong", 187, [Addr32, Reg8, Reg8], globals())

JStrictEqual = Instruction("JStrictEqual", 188, [Addr8, Reg8, Reg8], globals())
JStrictEqualLong = Instruction("JStrictEqualLong", 189, [Addr32, Reg8, Reg8], globals())

JStrictNotEqual = Instruction("JStrictNotEqual", 190, [Addr8, Reg8, Reg8], globals())
JStrictNotEqualLong = Instruction(
    "JStrictNotEqualLong", 191, [Addr32, Reg8, Reg8], globals()
)

Add32 = Instruction("Add32", 192, [Reg8, Reg8, Reg8], globals())

Sub32 = Instruction("Sub32", 193, [Reg8, Reg8, Reg8], globals())

Mul32 = Instruction("Mul32", 194, [Reg8, Reg8, Reg8], globals())

Divi32 = Instruction("Divi32", 195, [Reg8, Reg8, Reg8], globals())

Divu32 = Instruction("Divu32", 196, [Reg8, Reg8, Reg8], globals())

Loadi8 = Instruction("Loadi8", 197, [Reg8, Reg8, Reg8], globals())

Loadu8 = Instruction("Loadu8", 198, [Reg8, Reg8, Reg8], globals())

Loadi16 = Instruction("Loadi16", 199, [Reg8, Reg8, Reg8], globals())

Loadu16 = Instruction("Loadu16", 200, [Reg8, Reg8, Reg8], globals())

Loadi32 = Instruction("Loadi32", 201, [Reg8, Reg8, Reg8], globals())

Loadu32 = Instruction("Loadu32", 202, [Reg8, Reg8, Reg8], globals())

Store8 = Instruction("Store8", 203, [Reg8, Reg8, Reg8], globals())

Store16 = Instruction("Store16", 204, [Reg8, Reg8, Reg8], globals())

Store32 = Instruction("Store32", 205, [Reg8, Reg8, Reg8], globals())

_opcode_to_instruction: Dict[int, Instruction] = {v.opcode: v for v in _instructions}
_name_to_instruction: Dict[str, Instruction] = {v.name: v for v in _instructions}

_builtin_function_names: List[str] = [
    "Array.isArray",
    "Date.UTC",
    "Date.parse",
    "JSON.parse",
    "JSON.stringify",
    "Math.abs",
    "Math.acos",
    "Math.asin",
    "Math.atan",
    "Math.atan2",
    "Math.ceil",
    "Math.cos",
    "Math.exp",
    "Math.floor",
    "Math.hypot",
    "Math.imul",
    "Math.log",
    "Math.max",
    "Math.min",
    "Math.pow",
    "Math.round",
    "Math.sin",
    "Math.sqrt",
    "Math.tan",
    "Math.trunc",
    "Object.create",
    "Object.defineProperties",
    "Object.defineProperty",
    "Object.freeze",
    "Object.getOwnPropertyDescriptor",
    "Object.getOwnPropertyNames",
    "Object.getPrototypeOf",
    "Object.isExtensible",
    "Object.isFrozen",
    "Object.keys",
    "Object.seal",
    "String.fromCharCode",
    "silentSetPrototypeOf",
    "requireFast",
    "getTemplateObject",
    "ensureObject",
    "getMethod",
    "throwTypeError",
    "generatorSetDelegated",
    "copyDataProperties",
    "copyRestArgs",
    "arraySpread",
    "apply",
    "exportAll",
    "exponentiationOperator",
    "initRegexNamedGroups",
    "getOriginalNativeErrorConstructor",
    "spawnAsync",
]



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from typing import List, Dict, Set, Sequence, Union, Optional, Any
from io import BytesIO


def read_varint(
    buf: BytesIO,
) -> Optional[int]:  # From "decodeSLEB128" in the LLVH source code

    result = 0
    shift = 0
    byte = None

    while True:
        read = buf.read(1)
        if not read:
            break
        byte = read[0]
        result |= (byte & 0x7F) << shift
        shift += 7
        if not byte & 0x80:
            break

    if byte is None:
        return None

    # Sign-extend and convert if the result is negative
    if byte & 0x40:
        result = int.from_bytes(
            (((-1 << shift) | result) & 0xFFFFFFFF_FFFFFFFF).to_bytes(
                8, "little", signed=False
            ),
            "little",
            signed=True,
        )

    return result


def print_debug_info(buf: BytesIO, version: int):

    while True:
        function_index = read_varint(buf)
        if function_index is None:
            break
        current_line = read_varint(buf)
        current_column = read_varint(buf)

        print()
        print("Function index:", function_index)
        print("Start line:", current_line)
        print("Start column:", current_column)

        current_address = 0
        current_statement = 0

        while True:
            address_delta = read_varint(buf)
            if address_delta in (-1, None):
                break
            line_delta = read_varint(buf)
            column_delta = read_varint(buf)
            if version >= 94:
                scope_address = read_varint(buf)
                env_register = read_varint(buf)
            statement_delta = 0
            if line_delta & 1:
                statement_delta = read_varint(buf)
            line_delta >>= 1

            current_address += address_delta
            current_line += line_delta
            current_column += column_delta
            current_statement += statement_delta

            if version >= 94:
                print(
                    "  Address %d: Line %d - Column %d - Statement %d - Scope address %d - Env register %d"
                    % (
                        current_address,
                        current_line,
                        current_column,
                        current_statement,
                        scope_address,
                        env_register,
                    )
                )
            else:
                print(
                    "  Address %d: Line %d - Column %d - Statement %d"
                    % (current_address, current_line, current_column, current_statement)
                )

            # print('  Address delta:', address_delta)
            # print('  Line delta:', line_delta)
            # print('  Column delta:', column_delta)
            # print('  Statement delta:', statement_delta)

    print()


# WIP ..



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from typing import List, Set, Dict, Tuple, Optional, Sequence, Union, Any
from os.path import dirname, realpath
from dataclasses import dataclass
from sys import path
from re import sub
import sys

SCRIPT_DIR = dirname(realpath(__file__))
ROOT_DIR = realpath(SCRIPT_DIR + "/..")
PARSERS_DIR = realpath(ROOT_DIR + "/parsers")

path.insert(0, PARSERS_DIR)

from hbc_bytecode_parser import ParsedInstruction, get_builtin_functions
from hbc_file_parser import HBCReader


class HermesDecompiler:
    input_file: str
    output_file: Optional[str]

    calldirect_function_ids: Set[int]

    hbc_reader: HBCReader

    function_header: object  # For the function being transformed to dehydrated state
    indent_level: int = 0  # Used while producing decompilation output


# This should encompass the scope for a HBC assembly-level basic
# block (retranscribed as a for(;;) switch { case X: } case in
# the decompiled code, if not matching a particular NestingFrame
# object as defined below)
# This will provide the ability to construct a graph of
# the basic assembly blocks in the disassembled/decompiled
# output
class BasicBlock:
    start_address: int
    end_address: int

    # These flags should indicate whether we have
    # encountered cycling in
    # "graph_traversers/step2_visite_code_paths.py"
    # (which indicates the presence of a loop):
    may_be_cycling_anchor: bool = False
    may_be_cycling_target: bool = False

    # These flags delimite blocks that will mark
    # the code not continue into something
    # whichsoever
    is_unconditional_throw_anchor: bool = False
    is_unconditional_return_end: bool = False

    # These flags delimite blocks that have the
    # code to continue somewhere else
    is_unconditional_jump_anchor: bool = False
    is_yield_action_anchor: bool = False

    # These flags delimite blocks that
    # should have more than one branching
    # end
    is_conditional_jump_anchor: bool = False
    if_switch_action_anchor: bool = False

    anchor_instruction: Optional[ParsedInstruction] = None
    jump_targets_for_anchor: Optional[List[int]] = None

    # These attributes will be used to build
    # the decompiler's graph structure

    child_nodes: List["BasicBlock"]
    parent_nodes: List["BasicBlock"]

    error_handling_child_nodes: List["BasicBlock"]
    error_handling_parent_nodes: List["BasicBlock"]

    # These attributes should be added at steps 3+
    # of the graph traversal process (using
    # instruction pattern maching, etc)

    do_expr_at_begin: Optional["XXX_TODO"]
    if_expr_at_begin: Optional["XXX_TODO"]
    else_if_expr_at_begin: Optional["XXX_TODO"]
    else_expr_at_begin: Optional["XXX_TODO"]
    switch_expr_at_begin: Optional["XXX_TODO"]
    while_expr_at_begin: Optional["XXX_TODO"]
    for_expr_at_begin: Optional["XXX_TODO"]
    while_expr_at_end_matching_block: Optional["XXX_TODO"]
    closing_brace_at_end_matching_block: Optional["XXX_TODO"]

    # Whether this basic block should still be visible in the
    # decompiled code (switch this attribute to False when
    # a NestingFrame totally overlaps the concerned block)
    stay_visible: bool = True


# This should encompass the scope for a decompiled code-level
# nesting frame (switch, if, else if, for, while, do... block)
@dataclass
class NestedFrame:
    start_address: int
    end_address: int

    pass  # WIP


@dataclass
class Environment:
    parent_environment: Optional["Environment"] = None
    nesting_quantity: int = 0
    slot_index_to_varname: Dict[int, str] = None


class DecompiledFunctionBody:
    is_global: bool  # Is this the global function?
    function_name: str
    function_id: int
    function_object: object
    exc_handlers: "HBCExceptionHandlerInfo"

    try_starts: Dict[int, List[str]]
    try_ends: Dict[int, List[str]]
    catch_targets: Dict[int, List[str]]

    jump_anchors: Dict[
        int, ParsedInstruction
    ]  # Non-jump address associated with a jump/switch/yield instruction
    ret_anchors: Dict[
        int, ParsedInstruction
    ]  # Unreacheable address associated with a return instruction
    throw_anchors: Dict[
        int, ParsedInstruction
    ]  # Unreacheable address associated with a jump instruction
    jump_targets: Set[int]  # Jump target address

    is_closure: bool = False
    is_async: bool = False
    is_generator: bool = False

    # Closure variable naming-related attributes:
    parent_environment: Optional[Environment] = None
    environment_id: Optional[int] = None
    local_items: Dict[int, Environment] = {}  # {env_register: Environment}

    basic_blocks: List[BasicBlock]
    nested_frames: List[NestedFrame]

    statements: List["TokenString"]

    # Output decompiled code for this function to stdout
    def output_code(self, state: HermesDecompiler):
        output = ""

        if not self.is_global:  # Don't prototype the global function
            if self.is_async:
                output += "async "
            output += "function"
            if self.is_generator:
                output += "*"
            if not (self.is_closure or self.is_generator):
                assert self.function_name
                output += " " + self.function_name
            elif self.is_generator:
                output += " "
            output += "("
            output += ", ".join(
                "a" + str(index) for index in range(self.function_object.paramCount - 1)
            )
            # TODO: Handle function arguments properly otherwise
            output += ") {"
            if self.is_closure or self.is_generator:
                if self.function_name:
                    output += " // Original name: " + self.function_name
                    if self.environment_id is not None:
                        output += ", environment: r" + str(self.environment_id)
                elif self.environment_id is not None:
                    output += " // Environment: r" + str(self.environment_id)
            output += "\n"
            state.indent_level += 1

        nested_frame_starts = [
            nested_frame.start_address for nested_frame in self.nested_frames
        ]
        nested_frame_ends = [
            nested_frame.end_address for nested_frame in self.nested_frames
        ]

        basic_block_starts = [
            basic_block.start_address for basic_block in self.basic_blocks
        ]
        basic_block_ends = [
            basic_block.end_address for basic_block in self.basic_blocks
        ]

        if len(basic_block_starts) > 1:
            output += " " * (state.indent_level * 4)
            output += "_fun%d: for(var _fun%d_ip = 0; ; ) switch(_fun%d_ip) {\n" % (
                self.function_id,
                self.function_id,
                self.function_id,
            )
            state.indent_level += 1

        sys.stdout.write(output)
        output = ""

        # Process each statement, including nested functions

        for statement in self.statements:
            if statement.assembly:
                pos = statement.assembly[0].original_pos

                while pos in nested_frame_ends:
                    nested_frame_ends.pop(nested_frame_ends.index(pos))
                    state.indent_level -= 1
                    output += (" " * (state.indent_level * 4)) + "}\n"

                if len(basic_block_starts) > 1 and pos in basic_block_starts:
                    output += "case %d:" % pos

                    if pos in self.try_starts:
                        for label in self.try_starts[pos]:
                            output += " // %s" % label
                    if pos in self.try_ends:
                        for label in self.try_ends[pos]:
                            output += " // %s" % label
                    if pos in self.catch_targets:
                        for label in self.catch_targets[pos]:
                            output += " // %s" % label

                    output += "\n"

                while pos in nested_frame_starts:
                    nested_frame_starts.pop(nested_frame_starts.index(pos))
                    output += (" " * (state.indent_level * 4)) + "{\n"
                    state.indent_level += 1

                sys.stdout.write(output)
                output = ""

            if statement.tokens:
                sys.stdout.write(" " * (state.indent_level * 4))
                is_block = statement.tokens[:2] == [
                    RawToken("for"),
                    LeftParenthesisToken(),
                ]
                while statement.tokens:
                    op = statement.tokens.pop(0)
                    if isinstance(op, FunctionTableIndex):
                        op.closure_decompile(self)
                    elif isinstance(op, JumpNotCondition):
                        conditions = "".join(str(token) for token in statement.tokens)
                        statement.tokens = []
                        if conditions == "false":
                            sys.stdout.write(
                                "_fun%d_ip = %d; continue _fun%d"
                                % (
                                    self.function_id,
                                    op.target_address,
                                    self.function_id,
                                )
                            )
                        else:
                            is_block = True
                            sys.stdout.write("if(")
                            if "(" in conditions or " " in conditions:
                                sys.stdout.write("!(%s)" % conditions)
                            elif conditions[0] == "!":
                                sys.stdout.write(conditions[1:])
                            else:
                                sys.stdout.write("!" + conditions)
                            sys.stdout.write(
                                ") { _fun%d_ip = %d; continue _fun%d }"
                                % (
                                    self.function_id,
                                    op.target_address,
                                    self.function_id,
                                )
                            )
                    elif isinstance(op, JumpCondition):
                        conditions = "".join(str(token) for token in statement.tokens)
                        statement.tokens = []
                        if conditions == "true":
                            sys.stdout.write(
                                "_fun%d_ip = %d; continue _fun%d"
                                % (
                                    self.function_id,
                                    op.target_address,
                                    self.function_id,
                                )
                            )
                        else:
                            is_block = True
                            sys.stdout.write("if(")
                            sys.stdout.write(conditions)
                            sys.stdout.write(
                                ") { _fun%d_ip = %d; continue _fun%d }"
                                % (
                                    self.function_id,
                                    op.target_address,
                                    self.function_id,
                                )
                            )
                    else:
                        sys.stdout.write(str(op))
                if is_block:
                    sys.stdout.write("\n")
                else:
                    sys.stdout.write(";\n")

        if len(basic_block_starts) > 1:
            state.indent_level -= 1
            sys.stdout.write(" " * (state.indent_level * 4) + "}\n")

        if not self.is_global:
            state.indent_level -= 1
            sys.stdout.write(" " * (state.indent_level * 4) + "}")


@dataclass
class TokenString:
    tokens: List["Token"]
    assembly: ParsedInstruction


class Token:
    pass

    # def __str__(self):
    #     return repr(self).replace(self.__class__.__name__,
    #         sub('[^A-Z]', '', self.__class__.__name__), 1)


@dataclass
class ResumeGenerator(Token):
    result_out_reg: int
    return_bool_out_reg: int


@dataclass
class SaveGenerator(Token):
    address: int


@dataclass
class StartGenerator(Token):
    pass


@dataclass
class ReturnDirective(Token):
    def __str__(self):
        return "return "


@dataclass
class ThrowDirective(Token):
    def __str__(self):
        return "throw "


@dataclass
class LeftHandRegToken(Token):
    register: int

    def __str__(self):
        return "r%d" % self.register


@dataclass
class AssignmentToken(Token):
    def __str__(self):
        return " = "


@dataclass
class LeftParenthesisToken(Token):
    def __str__(self):
        return "("


@dataclass
class RightParenthesisToken(Token):
    def __str__(self):
        return ")"


@dataclass
class CatchBlockStart(Token):
    arg_register: int


@dataclass
class DotAccessorToken(Token):
    def __str__(self):
        return "."


@dataclass
class BindToken(Token):
    register: int

    def __str__(self):
        return ".bind(r%d)" % self.register


@dataclass
class RightHandRegToken(Token):
    register: int

    def __str__(self):
        return "r%d" % self.register


@dataclass
class GetEnvironmentToken(Token):
    register: int
    nesting_level: int


@dataclass
class LoadFromEnvironmentToken(Token):
    register: int
    slot_index: int


@dataclass
class NewEnvironmentToken(Token):
    register: int


@dataclass
class NewInnerEnvironmentToken(Token):
    dest_register: int
    parent_register: int
    number_of_slots: int


@dataclass
class SwitchImm(Token):
    value_reg: int
    jump_table_address: int
    default_jump_address: int
    unsigned_min_value: int
    unsigned_max_value: int


@dataclass
class StoreToEnvironment(Token):
    env_register: int
    slot_index: int
    value_register: int


@dataclass
class ForInLoopInit(Token):
    obj_props_register: int
    obj_register: int
    iter_index_register: int
    iter_size_register: int


@dataclass
class ForInLoopNextIter(Token):
    next_value_register: int
    obj_props_register: int
    obj_register: int
    iter_index_register: int
    iter_size_register: int


# Used to mark backward jumps (used in loops)
@dataclass
class JumpCondition(Token):
    target_address: int


# Used to mark forward jumps
@dataclass
class JumpNotCondition(Token):
    target_address: int


@dataclass
class FunctionTableIndex(Token):
    function_id: int
    state: HermesDecompiler
    environment_id: Optional[int] = None

    is_closure: bool = False
    is_builtin: bool = False
    is_generator: bool = False
    is_async: bool = False

    parent_environment: Optional[Environment] = None

    def closure_decompile(self, parent_closure: DecompiledFunctionBody):

        # Is this a nested closure?
        # If yes, print its decompiled code
        if (self.is_closure or self.is_generator) and not self.is_builtin:
            import hbc_decompiler

            hbc_decompiler.decompile_function(
                self.state,
                self.function_id,
                parent_environment=self.parent_environment,
                environment_id=self.environment_id,
                is_closure=self.is_closure,
                is_generator=self.is_generator,
                is_async=self.is_async,
            )
            return

        # Is this a know builtin or named function?
        name = self._name_if_any()
        if name:
            sys.stdout.write(name)
            return

        # Is this anything else?
        sys.stdout.write(
            "<Function #%d: %s>"
            % (
                self.function_id,
                ", ".join(
                    "%s: %s" % (key, value)
                    for key, value in self.__dict__.items()
                    if key != "state" and value
                ),
            )
        )

    def _name_if_any(self) -> Optional[str]:

        if self.is_builtin:
            builtin_functions = get_builtin_functions(
                self.state.hbc_reader.parser_module
            )
            return builtin_functions[self.function_id]

        else:
            return (
                self.state.hbc_reader.strings[
                    self.state.hbc_reader.function_headers[
                        self.function_id
                    ].functionName
                ]
                or None
            )


@dataclass
class RawToken(Token):
    token: str

    def __str__(self):
        return self.token


# WIP .. .



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from typing import List, Union, Any, Sequence, Dict, Set, Optional
from dataclasses import dataclass
from enum import IntEnum, IntFlag
from struct import unpack
from io import BytesIO

"""
    The SerializedLiteral format is used to store structured
    arrays of string references and literals within the Hermes
    array buffer, object key buffer, object value buffer
    arrays.
    
    It is parsed here:
    
        https://github.com/facebook/hermes/blob/v0.12.0/lib/BCGen/HBC/BytecodeDisassembler.cpp#L285
    
  OS << "Object Key Buffer:\n";
  while ((size_t)keyInd < objKeyBuffer.size()) {
    std::pair<int, SLG::TagType> keyTag =
        checkBufferTag(objKeyBuffer.data() + keyInd);
    keyInd += (keyTag.first > 0x0f ? 2 : 1);
    for (int i = 0; i < keyTag.first; i++) {
      OS << SLPToString(keyTag.second, objKeyBuffer.data(), &keyInd) << "\n";
    }
  }

    
    Strings are referred to as indexes of the global string,
    which are stored over either 8, 16 or 32 bits each:
    
        int ind =
            isKeyBuffer ? BMGen_.getIdentifierID(str) : BMGen_.getStringID(str);
    
    The binary form of this format is TLV-oriented, see:
    
    https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/BCGen/HBC/SerializedLiteralGenerator.h
    https://github.com/facebook/hermes/blob/v0.12.0/lib/BCGen/HBC/SerializedLiteralParserBase.cpp
    https://github.com/facebook/hermes/blob/v0.12.0/lib/BCGen/HBC/SerializedLiteralGenerator.cpp
    https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/BCGen/HBC/SerializedLiteralParserBase.h
    https://github.com/facebook/hermes/blob/v0.12.0/lib/VM/SerializedLiteralParser.cpp
    https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/VM/SerializedLiteralParser.h
"""


class TagType(IntEnum):
    NullTag = 0
    TrueTag = 1
    FalseTag = 2
    NumberTag = 3
    LongStringTag = 4
    ShortStringTag = 5
    ByteStringTag = 6
    IntegerTag = 7


@dataclass
class SLPValue:
    tag_type: TagType

    value: object


@dataclass
class SLPArray:
    items: List[SLPValue]

    def to_strings(self, string_table: List[str]) -> List[str]:

        stringified_items: List[str] = []
        for item in self.items:
            if item.tag_type == TagType.NullTag:
                string = "null"
            elif item.tag_type == TagType.TrueTag:
                string = "true"
            elif item.tag_type == TagType.FalseTag:
                string = "false"
            elif item.tag_type == TagType.NumberTag:
                string = str(item.value)
            elif item.tag_type in (
                TagType.LongStringTag,
                TagType.ShortStringTag,
                TagType.ByteStringTag,
            ):
                string = repr(string_table[item.value])
            elif item.tag_type == TagType.IntegerTag:
                string = str(item.value)
            else:
                raise ValueError

            stringified_items.append(string)

        return stringified_items


def unpack_slp_array(data: bytes, num_items: int) -> SLPArray:
    data = BytesIO(data)

    items = []

    while len(items) < num_items:
        values = []

        next_tag = data.read(1)
        if not next_tag:
            break
        tag_type = TagType((next_tag[0] >> 4) & 0b111)
        if next_tag[0] >> 7 == 1:  # Extended length flag?
            length = ((next_tag[0] & 0b1111) << 8) | data.read(1)[0]
        else:
            length = next_tag[0] & 0b1111

        for item in range(length):
            if tag_type == TagType.NullTag:
                values.append(None)
            elif tag_type == TagType.TrueTag:
                values.append(True)
            elif tag_type == TagType.FalseTag:
                values.append(False)
            elif tag_type == TagType.NumberTag:
                values.append(unpack("<d", data.read(8))[0])
            elif tag_type == TagType.LongStringTag:
                values.append(int.from_bytes(data.read(4), "little"))
            elif tag_type == TagType.ShortStringTag:
                values.append(int.from_bytes(data.read(2), "little"))
            elif tag_type == TagType.ByteStringTag:
                values.append(int.from_bytes(data.read(1), "little"))
            elif tag_type == TagType.IntegerTag:
                values.append(int.from_bytes(data.read(4), "little"))
            else:
                raise ValueError

        items.extend(SLPValue(tag_type, value) for value in values)

    return SLPArray(items[:num_items])



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from typing import List, Dict, Set, Optional, Any, Sequence
from os.path import dirname, realpath
from argparse import ArgumentParser
from json import dumps
import sys

DISASSEMBLY_DIR = dirname(realpath(__file__))
ROOT_DIR = realpath(DISASSEMBLY_DIR + "/..")
PARSERS_DIR = realpath(ROOT_DIR + "/parsers")
sys.path.insert(0, PARSERS_DIR)

from hbc_bytecode_parser import parse_hbc_bytecode
from hbc_file_parser import HBCReader, StringKind


def do_disassemble(input_file: str):

    with open(input_file, "rb") as file_descriptor:
        hbc_reader = HBCReader()

        hbc_reader.read_whole_file(file_descriptor)

        """
        output_json_data : Dict[int, dict] = {}
        
        identifier_count = 0
        for string_count, (string_kind, string) in enumerate(zip(hbc_reader.string_kinds, hbc_reader.strings)):
            
            obj = hbc_reader.small_string_table[string_count]
            
            output_json_data[string_count] = {
                'type': {
                    StringKind.String: 'string',
                    StringKind.Identifier: 'identifier',
                    StringKind.Predefined: 'predefined'
                }[string_kind],
                'value': string,
                'is_utf16': bool(obj.isUTF16)
            }
        
        print('{')
        print(',\n'.join(
            '    %d: %s' % (key, dumps(value))
            for key, value in sorted(output_json_data.items())
        ))
        print('}')
        """
        # WIP ..²

        # TODO : Print the JSON data generated above, one line per entry I guess?

        for function_count, function_header in enumerate(hbc_reader.function_headers):
            # pretty_print_structure(function_header)
            exception_info = ""
            if function_header.hasExceptionHandler:
                exception_data = hbc_reader.function_id_to_exc_handlers[function_count]
                exception_info = "\n  [Exception handlers:"
                for exception_item in exception_data:
                    exception_info += " [start=" + hex(exception_item.start) + ", "
                    exception_info += "end=" + hex(exception_item.end) + ", "
                    exception_info += "target=" + hex(exception_item.target) + "]"
                exception_info += " ]"

            debug_info = ""
            if function_header.hasDebugInfo:
                debug_data = hbc_reader.function_id_to_debug_offsets[function_count]
                debug_info = "\n  [Debug offsets: "
                debug_info += "source_locs=" + hex(debug_data.source_locations) + ", "
                debug_info += "scope_desc_data=" + hex(debug_data.scope_desc_data) + "]"

            print(
                '=> [Function #%d "%s" of %d bytes]: %d params, frame size=%d, env size=%d, read index sz=%d, write index sz=%d, strict=%r, exc handler=%r, debug info=%r  @ offset 0x%08x%s%s'
                % (
                    function_count,
                    hbc_reader.strings[function_header.functionName],
                    function_header.bytecodeSizeInBytes,
                    function_header.paramCount,
                    function_header.frameSize,
                    function_header.environmentSize,
                    function_header.highestReadCacheIndex,
                    function_header.highestWriteCacheIndex,
                    function_header.strictMode,
                    function_header.hasExceptionHandler,
                    function_header.hasDebugInfo,
                    function_header.offset,
                    exception_info,
                    debug_info,
                )
            )

            print()
            print("Bytecode listing:")
            print()
            for instruction in parse_hbc_bytecode(function_header, hbc_reader):
                print("==>", repr(instruction))
            print()
            print()
            print("=" * 15)
            print()

            # Safety checks:
            assert function_header.unused == 0 and function_header.paramCount < 100


def main():

    # TODO : Make an actual CLI with extra options here ..

    args = ArgumentParser()

    args.add_argument("input_file")
    args.add_argument("output_file", nargs="?")  # , default = '/dev/stdout'

    args = args.parse_args()

    if args.output_file:
        stdout = sys.stdout
        with open(args.output_file, "w", encoding="utf-8") as sys.stdout:
            do_disassemble(args.input_file)
        sys.stdout = stdout

        print()
        print('[+] Disassembly output wrote to "%s"' % args.output_file)
        print()

    else:
        sys.stdout.reconfigure(encoding="utf-8")
        do_disassemble(args.input_file)


if __name__ == "__main__":
    main()
